###### Trainer ##########
import warnings
from typing import Optional, Union
from types import SimpleNamespace
from collections import defaultdict
import os
import time
import matplotlib.pyplot as plt

import numpy as np
import random
import scanpy as sc
from anndata import concat

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.data import Data
from torch_geometric.data import InMemoryDataset
from torch_geometric.sampler import NeighborSampler
from torch_geometric.loader import NeighborLoader, GraphSAINTEdgeSampler, NodeLoader
import torch_geometric.transforms as T

from .._settings import settings
import matplotlib
from matplotlib import pyplot as plt
from ._utils import plot_alluvial
# from ._utils import extract_subgraph
from ._utils import split_data
from ._tools import EarlyStopping, print_progress
from ._loss import VGAE_loss, InstanceLoss, ClusterLoss, mmd_loss_calc, unbalanced_ot, cosine_sim
from .prepare_Data import UserDataset
from .Garfield_net import Garfield
from .transfer_base import BaseMixin, SurgeryMixin
from .transfer_anno import weighted_knn_trainer, weighted_knn_transfer

# from torch.optim.lr_scheduler import StepLR
# from scheduler import CosineAnnealingWarmRestarts
# GAMMA=0.9
warnings.filterwarnings("ignore", category=UserWarning, module="torch_geometric")

class GarfieldTrainer(BaseMixin, SurgeryMixin):
    """
    Garfield model trainer.
    """
    def __init__(self, gf_params):
        """
        :param args: Arguments object.
        """
        if gf_params is None:
            gf_params = settings.gf_params.copy()
        else:
            assert isinstance(gf_params, dict), \
                "`gf_params` must be dict"

        self.args = SimpleNamespace(**gf_params)
        ## data preprocess parameter
        self.data_dir_ = self.args.data_dir
        self.project_name_ = self.args.project_name
        self.adata_list_ = self.args.adata_list
        self.profile_ = self.args.profile
        self.projection_ = self.args.projection
        self.data_type_ = self.args.data_type
        self.sub_data_type_ = self.args.sub_data_type
        self.weight_ = self.args.weight
        self.graph_const_method_ = self.args.graph_const_method
        self.genome_ = self.args.genome
        self.sample_col_ = self.args.sample_col
        self.filter_cells_rna_ = self.args.filter_cells_rna
        self.min_features_ = self.args.min_features
        self.min_cells_ = self.args.min_cells
        self.keep_mt_ = self.args.keep_mt
        self.normalize_ = self.args.normalize
        self.target_sum_ = self.args.target_sum
        self.used_hvg_ = self.args.used_hvg
        self.used_scale_ = self.args.used_scale
        self.single_n_top_genes_ = self.args.single_n_top_genes
        self.rna_n_top_features_ = self.args.rna_n_top_features
        self.atac_n_top_features_ = self.args.atac_n_top_features
        self.metacell_size_ = self.args.metacell_size
        self.metacell_ = self.args.metacell
        self.n_pcs_ = self.args.n_pcs
        self.n_neighbors_ = self.args.n_neighbors
        self.metric_ = self.args.metric
        self.svd_solver_ = self.args.svd_solver
        self.method_ = self.args.method
        self.resolution_tol_ = self.args.resolution_tol
        self.leiden_runs_ = self.args.leiden_runs
        self.leiden_seed_ = self.args.leiden_seed
        self.verbose_ = self.args.verbose

        ## model parameter
        self.epochs_ = self.args.epochs
        self.hidden_dims_ = self.args.hidden_dims
        self.bottle_neck_neurons_ = self.args.bottle_neck_neurons
        self.num_heads_ = self.args.num_heads
        self.dropout_ = self.args.dropout
        self.concat_ = self.args.concat
        self.svd_q_ = self.args.svd_q
        self.used_edge_weight_ = self.args.used_edge_weight
        self.used_mmd_ = self.args.used_mmd
        self.used_DSBN_ = self.args.used_DSBN
        self.used_recon_exp_ = self.args.used_recon_exp
        self.conv_type_ = self.args.conv_type
        self.cluster_num_ = self.args.cluster_num
        self.gnn_layer_ = self.args.gnn_layer
        self.instance_temperature_ = self.args.instance_temperature
        self.cluster_temperature_ = self.args.cluster_temperature
        self.mmd_temperature_ = self.args.mmd_temperature
        self.l2_reg_ = self.args.l2_reg
        self.patience_ = self.args.patience
        self.gradient_clipping_ = self.args.gradient_clipping
        self.learning_rate_ = self.args.learning_rate
        self.weight_decay_ = self.args.weight_decay

        ## data loader
        self.loader_type_ = self.args.loader_type
        self.num_neighbors_ = self.args.num_neighbors
        self.batch_size_ = self.args.batch_size
        self.test_split_ = self.args.test_split
        self.val_split_ = self.args.val_split

        self.monitor_only_val_losses_ = self.args.monitor_only_val_losses
        self.outdir_ = self.args.outdir
        self.impute_ = self.args.impute

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        # self.device = torch.device("cpu")
        seed = 2024
        random.seed(seed)
        os.environ['PYTHONHASHSEED'] = str(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

        self.load_dataset()
        self.setup_model()

    # Load dataset
    def load_dataset(self):
        """
        Loading and processing dataset.
        """
        print("\nPreparing dataset...\n")
        self.training_graphs = UserDataset(root="{}/".format(self.data_dir_),
                                           project_name=self.project_name_,
                                           adata_list=self.adata_list_,
                                           profile=self.profile_,
                                           projection=self.projection_,
                                           data_type=self.data_type_,
                                           sub_data_type=self.sub_data_type_,
                                           weight=self.weight_,
                                           graph_const_method=self.graph_const_method_,
                                           genome=self.genome_,
                                           sample_col=self.sample_col_,
                                           filter_cells_rna=self.filter_cells_rna_,
                                           min_features=self.min_features_,
                                           min_cells=self.min_cells_,
                                           keep_mt=self.keep_mt_,
                                           normalize=self.normalize_,
                                           target_sum=self.target_sum_,
                                           used_hvg=self.used_hvg_,
                                           used_scale=self.used_scale_,
                                           single_n_top_genes=self.single_n_top_genes_,
                                           rna_n_top_features=self.rna_n_top_features_,
                                           atac_n_top_features=self.atac_n_top_features_,
                                           metacell_size=self.metacell_size_,
                                           metacell=self.metacell_,
                                           n_pcs=self.n_pcs_,
                                           n_neighbors=self.n_neighbors_,
                                           metric=self.metric_,
                                           svd_solver=self.svd_solver_,
                                           method=self.method_,
                                           resolution_tol=self.resolution_tol_,
                                           leiden_runs=self.leiden_runs_,
                                           leiden_seed=self.leiden_seed_,
                                           verbose=self.verbose_
                                           )
        self.merged_adata = self.training_graphs.merged_adata
        self.num_classes = self.training_graphs.num_classes ## batch number
        self.number_of_genes = self.training_graphs.num_features

        ## 划分 train_test_split_edges 构造Data对象
        self.training_graphs = self.testing_graphs = self.training_graphs[0]

    def setup_model(self):
        """
        Creating a Garfield.
        """
        self.model = Garfield(
            self.number_of_genes,
            self.hidden_dims_,
            self.bottle_neck_neurons_,
            self.num_classes,
            self.num_heads_,
            self.dropout_,
            self.concat_,
            self.svd_q_,
            self.used_edge_weight_,
            self.used_DSBN_,
            self.used_recon_exp_,
            self.conv_type_,
            self.cluster_num_,
            self.gnn_layer_
        )
        self.model = self.model.to(self.device)
        self.is_trained_ = False
        self.epoch = -1
        self.n_epochs = None
        self.training_time = 0
        self.logs = defaultdict(list) ## logs

        if self.outdir_ is not None:
            outdir = self.outdir_
        else:
            outdir = self.data_dir_
        os.makedirs(os.path.join(outdir, 'checkpoint'), exist_ok=True)

        self.criterion_instance = InstanceLoss(self.instance_temperature_, self.device).to(
            self.device)
        self.criterion_cluster = ClusterLoss(self.cluster_num_, self.cluster_temperature_, self.device).to(
            self.device)
        self.early_stopping = EarlyStopping(patience=self.patience_, checkpoint_file=os.path.join(outdir, 'checkpoint/model.pt'))

    # def save(self):
    #     """
    #     Saving model.
    #     """
    #     if self.args.outdir is not None:
    #         outdir = self.args.outdir
    #     else:
    #         outdir = self.args.data_dir
    #     torch.save(self.model.state_dict(), os.path.join(outdir, 'checkpoint/model.pt'))  # self.args.save
    #     print(f"Model is saved under {os.path.join(outdir, 'checkpoint/model.pt')}.")

    def load(self):
        """
        Loading model.
        """
        if self.outdir_ is not None:
            outdir = self.outdir_
        else:
            outdir = self.data_dir_

        self.model.load_state_dict(torch.load(os.path.join(outdir, 'checkpoint/model.pt')))
        print(f"Model is loaded from {os.path.join(outdir, 'checkpoint/model.pt')}.")

    def create_batches(self, is_train=True):
        """
        Creating batches from the training graph list.
        :return batches: training or testing loaders.
        """
        # Split into train/val/test
        # self.training_graphs.train_mask, self.training_graphs.val_mask, self.training_graphs.test_mask = split_data(self.training_graphs.y.size(0), self.val_split_, self.test_split_)

        if is_train:
            if self.loader_type_ == "neighbor":
                if self.num_neighbors_ is None:
                    self.num_neighbors_ = [3, 3]
                loader = NeighborLoader(self.training_graphs,
                                        num_neighbors=self.num_neighbors_,
                                        batch_size=self.batch_size_,
                                        input_nodes=torch.arange(self.training_graphs.num_nodes),
                                        shuffle=True)
            elif self.loader_type_ == "graphsaint":
                # loader = GraphSAINTRandomWalkSampler(data, batch_size = batch_size, walk_length = num_layers)
                loader = GraphSAINTEdgeSampler(self.training_graphs, batch_size=self.batch_size_, num_steps=16)
            else:
                raise NotImplementedError
        else:
            node_sampler = NeighborSampler(self.testing_graphs, num_neighbors=[0])  # num_neighbors=[0]表示不采样任何邻居
            # 设置 DataLoader
            loader = NodeLoader(
                self.testing_graphs,
                batch_size=self.batch_size_,
                shuffle=False,
                node_sampler=node_sampler
            )
        return loader

    def process_batch(self, batch_data, device):
        """
        Forward pass with a data.
        :param data: Data that is essentially pair of batches, for source and target graphs.
        :return loss: Loss on the data.
        """
        # Dont update any weight on first layers
        if self.model.freeze:
            for name, module in self.model.named_modules():
                if isinstance(module, nn.BatchNorm1d):
                    if not module.weight.requires_grad:
                        module.affine = False
                        module.track_running_stats = False

        self.optimizer.zero_grad()
        ## calculate loss
        loss = self.calculate_losses(batch_data, device)
        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.gradient_clipping_)
        self.optimizer.step()

    def validate_batch(self, val_data, test_data, device):
        # Get Train Epoch Logs
        for key in self.iter_logs:
            self.logs["epoch_" + key].append(np.array(self.iter_logs[key]).mean())

        # Validate Model
        vali_loss = self.validate(val_data, device)

        # AUC & precision
        self.test_metrics(test_data)

        # Monitor Logs
        if self.monitor_only_val_losses_:
            print_progress(self.epoch, self.logs, self.n_epochs, only_val_losses=True)
        else:
            print_progress(self.epoch, self.logs, self.n_epochs, only_val_losses=False)

        return vali_loss

    def calculate_losses(self, batch_data, device):
        """
        Calculate various losses for the batch data using the model.

        Parameters:
        - batch_data: The batch of data.
        - device: The device tensors are moved to.

        Returns:
        - all computed losses.
        """
        batch_mat = batch_data.x
        cell_batch = batch_data.y

        if self.used_recon_exp_:
            z, z_1, z_2, c_1, c_2, mu, logstd, recon_features = self.model(batch_data)
        else:
            z, z_1, z_2, c_1, c_2, mu, logstd = self.model(batch_data)

        # loss calculations
        # 1.VGAE ELBO loss
        vgae_loss = VGAE_loss(self.model.VGAE, z, mu, logstd, batch_data.pos_edge_label_index).to(device)

        # 2.Reconstruction loss
        if self.used_recon_exp_:
            ## total expr loss
            recon_loss = F.mse_loss(batch_mat, recon_features) * batch_mat.size(-1)

        # 3.Contrastive losses
        loss_instance = self.criterion_instance(z_1, z_2)
        loss_cluster = self.criterion_cluster(c_1, c_2)
        cl_loss = loss_instance + loss_cluster

        # 4.MMD loss
        mmd_loss = torch.tensor(0.0).to(device)
        if self.used_mmd_:
            cell_batch = cell_batch.detach().cpu()
            unique_groups, group_indices = np.unique(cell_batch, return_inverse=True)
            grouped_z_cell = {group: z[group_indices == i] for i, group in enumerate(unique_groups)}
            group_labels = list(unique_groups)
            num_groups = len(group_labels)

            for i in range(num_groups):
                for j in range(i + 1, num_groups):
                    z_i = grouped_z_cell[group_labels[i]]
                    z_j = grouped_z_cell[group_labels[j]]
                    mmd_loss_tmp = mmd_loss_calc(z_i, z_j)
                    mmd_loss += mmd_loss_tmp * z.size(0)

        # 5.Regularization loss
        regu_loss = torch.tensor(0.0).to(device)
        for param in self.model.parameters():
            regu_loss += param.norm(2).square()

        ### total loss
        loss = vgae_loss + recon_loss + cl_loss + regu_loss * self.l2_reg_ + mmd_loss * self.mmd_temperature_

        self.iter_logs["total_loss"].append(loss.item())
        self.iter_logs["vgae_loss"].append(vgae_loss.item())
        self.iter_logs["instance_loss"].append(loss_instance.item())
        self.iter_logs["cluster_loss"].append(loss_cluster.item())
        if self.used_recon_exp_:
            self.iter_logs["recon_loss"].append(recon_loss.item())
        if self.used_mmd_:
            self.iter_logs["mmd_loss"].append(mmd_loss.item())
        self.iter_logs["regu_loss"].append(regu_loss.item())

        return loss

    def fit(self):
        """
        Training a model.
        """
        print("\nModel training.\n")
        begin = time.time()
        params = filter(lambda p: p.requires_grad, self.model.parameters())
        self.optimizer = torch.optim.Adam(  # AdamW
            params,
            lr=self.learning_rate_,
            weight_decay=self.weight_decay_,
        )
        # self.scheduler = StepLR(self.optimizer, step_size=1, gamma=GAMMA)
        # self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
        #     self.optimizer, mode="min", factor=0.5, patience=5, min_lr=1e-6
        # )
        self.model.train()

        self.n_epochs = num_epochs = self.epochs_
        for self.epoch in range(num_epochs):
            self.iter_logs = defaultdict(list)
            loader = self.create_batches(is_train=True)
            for index, batch in enumerate(loader):
                if self.test_split_ > 0 and self.val_split_ > 0:
                    test_split = self.test_split_
                    val_split = self.val_split_
                    # Can set validation ratio
                    transform = T.RandomLinkSplit(num_val=val_split, num_test=test_split, is_undirected=True,
                                                  add_negative_train_samples=False, split_labels=True)
                    train_data, val_data, test_data = transform(batch)
                else:
                    train_data = batch.to(self.device)
                    # 直接使用 edge_index 作为 pos_edge_label_index
                    train_data.pos_edge_label_index = train_data.edge_index
                    val_data = test_data = train_data

                train_data = train_data.to(self.device)
                val_data = val_data.to(self.device)
                test_data = test_data.to(self.device)
                ## Loss Calculation(training mode)
                self.process_batch(train_data, self.device)
                torch.cuda.empty_cache()

                # Validation of Model, Monitoring, Early Stopping
                vali_loss = self.validate_batch(val_data, test_data, self.device)
                vali_loss = vali_loss.detach().cpu() #.numpy()
                # 在每个epoch结束时调用scheduler.step()来更新学习率
                # self.scheduler.step()

            ## early_stopping
            self.early_stopping(vali_loss, self.model)
            if self.early_stopping.early_stop:
                print('\n')
                print('EarlyStopping: run {} epoch'.format(self.epoch + 1))
                break

        self.is_trained_ = True
        self.training_time += (time.time() - begin)
        ## 保存模型
        # self.save()

    @classmethod
    def _get_init_params_from_dict(cls, dct):
        init_params = {
            # Input options
            'data_dir': dct['data_dir_'],
            'project_name': dct['project_name_'],
            'adata_list': dct['adata_list_'],
            'profile': dct['profile_'],
            'data_type': dct['data_type_'],
            'weight': dct['weight_'],
            'genome': dct['genome_'],
            'sample_col': dct['sample_col_'],

            ## whether to use metacell mode
            'metacell': dct['metacell_'],
            'metacell_size': dct['metacell_size_'],
            'single_n_top_genes': dct['single_n_top_genes_'],
            'n_pcs': dct['n_pcs_'],

            # Preprocessing options
            'filter_cells_rna': dct['filter_cells_rna_'],
            'min_features': dct['min_features_'],
            'min_cells': dct['min_cells_'],
            'keep_mt': dct['keep_mt_'],
            'normalize': dct['normalize_'],
            'target_sum': dct['target_sum_'],
            'used_hvg': dct['used_hvg_'],
            'used_scale': dct['used_scale_'],
            'rna_n_top_features': dct['rna_n_top_features_'],
            'atac_n_top_features': dct['atac_n_top_features_'],
            'n_neighbors': dct['n_neighbors_'],
            'svd_solver': dct['svd_solver_'],
            'method': dct['method_'],
            'metric': dct['metric_'],
            'resolution_tol': dct['resolution_tol_'],
            'leiden_runs': dct['leiden_runs_'],
            'leiden_seed': dct['leiden_seed_'],
            'verbose': dct['verbose_'],

            # Model options
            'gnn_layer': dct['gnn_layer_'],
            'conv_type': dct['conv_type_'],
            'hidden_dims': dct['hidden_dims_'],
            'bottle_neck_neurons': dct['bottle_neck_neurons_'],
            'svd_q': dct['svd_q_'],
            'cluster_num': dct['cluster_num_'],
            'num_heads': dct['num_heads_'],
            'concat': dct['concat_'],
            'used_edge_weight': dct['used_edge_weight_'],
            'used_recon_exp': dct['used_recon_exp_'],
            'used_DSBN': dct['used_DSBN_'],
            'used_mmd': dct['used_mmd_'],
            'test_split': dct['test_split_'],
            'val_split': dct['val_split_'],
            'batch_size': dct['batch_size_'],
            'loader_type': dct['loader_type_'],
            'num_neighbors': dct['num_neighbors_'],
            'epochs': dct['epochs_'],
            'dropout': dct['dropout_'],
            'mmd_temperature': dct['mmd_temperature_'],
            'instance_temperature': dct['instance_temperature_'],
            'cluster_temperature': dct['cluster_temperature_'],
            'l2_reg': dct['l2_reg_'],
            'patience': dct['patience_'],
            'monitor_only_val_losses': dct['monitor_only_val_losses_'],
            'gradient_clipping': dct['gradient_clipping_'],
            'learning_rate': dct['learning_rate_'],
            'weight_decay': dct['weight_decay_'],

            # Other options
            'projection': dct['projection_'],
            'impute': dct['impute_'],
            'outdir': dct['outdir_'],
            'load': False
        }

        return init_params

    def label_transfer(self,
                       ref_adata,
                       ref_adata_emb,
                       query_adata,
                       query_adata_emb,
                       ref_adata_obs,
                       label_keys,
                       n_neighbors=50,
                       threshold=1,
                       pred_unknown=False,
                       mode="package"):
        knn_transformer = weighted_knn_trainer(
            train_adata=ref_adata,
            train_adata_emb=ref_adata_emb,  # location of our joint embedding
            n_neighbors=n_neighbors,
        )

        labels, uncert = weighted_knn_transfer(
            query_adata=query_adata,
            query_adata_emb=query_adata_emb,  # location of our embedding, query_adata.X in this case
            label_keys=label_keys,  # (start of) obs column name(s) for which to transfer labels
            knn_model=knn_transformer,
            ref_adata_obs=ref_adata_obs,
            threshold=threshold,
            pred_unknown=pred_unknown,
            mode=mode
        )

        # 定义列名的映射
        if pred_unknown:
            rename_mapping_labels = {label_keys: f"transferred_{label_keys}_filtered"}
        else:
            rename_mapping_labels = {label_keys: f"transferred_{label_keys}_unfiltered"}
        rename_mapping_uncert = {label_keys: f"transferred_{label_keys}_uncert"}

        # 重命名列并加入到 'query_adata.obs'
        query_adata.obs = query_adata.obs.join(
            labels.rename(columns=rename_mapping_labels)
        )
        # 重命名列并加入到 'query_adata.obs'
        query_adata.obs = query_adata.obs.join(
            uncert.rename(columns=rename_mapping_uncert)
        )

        return query_adata

    @torch.no_grad()
    def validate(self, val_data, device):
        self.model.eval()
        self.iter_logs = defaultdict(list)

        ## validation loss
        vali_loss = self.calculate_losses(val_data, device)

        # Get Validation Logs
        for key in self.iter_logs:
            self.logs["val_" + key].append(np.array(self.iter_logs[key]).mean())

        self.model.train()

        return vali_loss

    @torch.no_grad()
    def test_metrics(self, test_data):
        self.model.eval()

        # Calculate test metrics
        roc_score, ap_score, acc, f1 = self.model.VGAE.single_test(test_data)

        ## roc_auc, ap info
        self.logs["test_" + 'roc_auc'].append(np.array(roc_score).mean())
        self.logs["test_" + 'precision'].append(np.array(ap_score).mean())
        self.logs["test_" + 'accuracy'].append(np.array(acc).mean())
        self.logs["test_" + 'f1_score'].append(np.array(f1).mean())

        self.model.train()

    ## Loss curve
    def plot_losses_curve(self, title="Training Losses Curve", show=True, save=False, dir_path=None):
        if save:
            show = False
            if dir_path is None:
                save = False

        fig = plt.figure()
        elbo_train = self.logs["epoch_total_loss"]
        elbo_test = self.logs["val_total_loss"]
        x = np.linspace(0, len(elbo_train), num=len(elbo_train))
        plt.plot(x, elbo_train, label="Train")
        plt.plot(x, elbo_test, label="Validate")
        plt.ylim(min(elbo_test) - 50, max(elbo_test) + 50)
        plt.legend()
        plt.title(title)
        if save:
            plt.savefig(f'{dir_path}.png', bbox_inches='tight')
        if show:
            plt.show()
        plt.clf()

    ## 获取隐变量
    @torch.no_grad()
    def get_latent(self):
        self.model.eval()
        print('eval mode')
        print('Perform get_latent for cells via mini-batch mode')

        cmu = None
        recon = None
        Data_test = self.create_batches(is_train=False)

        for batch in Data_test:
            batch = batch.to(self.device)
            if self.used_recon_exp_:
                if self.impute_:
                    print('impute mode')
                    _, x_recon = self.model.encodeBatch(batch)  # mu, x_recon
                    ## expression profile
                    if recon is None:
                        recon = x_recon.clone().detach().cpu()
                    else:
                        recon = torch.cat([recon, x_recon.cpu()], dim=0)
                else:
                    mu, _ = self.model.encodeBatch(batch)  # mu, x_recon
            else:
                mu = self.model.encodeBatch(batch)

            ## latent
            if cmu is None:
                cmu = mu.clone().detach().cpu()
            else:
                cmu = torch.cat([cmu, mu.cpu()], dim=0)

        latent = cmu.detach().numpy()

        if self.impute_:
            exp_mat = recon.detach().numpy()
            return latent, exp_mat
        else:
            return latent

    @torch.no_grad()
    def get_latent_representation(self):
        if self.impute_:
            self.merged_adata.obsm['X_gf'], self.merged_adata.layers['X_impute'] = self.get_latent()
        else:
            self.merged_adata.obsm['X_gf'] = self.get_latent()
        # sc.pp.neighbors(self.merged_adata, use_rep='X_gf')
        # 保存数据
        if self.outdir_ is not None:
            outdir = self.outdir_
        else:
            outdir = self.data_dir_
        self.merged_adata.write(os.path.join(outdir, 'checkpoint/adata_ref.h5ad'), compression='gzip')

        return self.merged_adata

    def plot_latent(self,
                    color_key=['celltype', 'batch'],
                    show=True,
                    save=False,
                    dir_path=None,
                    n_neighbors=8,
                    in_one=True,
                    colors=None):
        """
        if save:
            if dir_path is None:
                name = 'garfield_latent.png'
            else:
                name = f'{dir_path}.png'
        else:
            name = False
        """
        if save:
            show = False
            if dir_path is None:
                dir_path = 'garfield_latent'
        sc.pp.neighbors(self.merged_adata, n_neighbors=n_neighbors, use_rep='X_gf')
        sc.tl.umap(self.merged_adata)
        if in_one:
            sc.pl.umap(self.merged_adata,
                       color=color_key,
                       ncols=2,
                       frameon=False,
                       wspace=0.6,
                       show=show,
                       save=f'{dir_path}_complete.png' if dir_path else None)
        else:
            if 'celltype' in color_key:
                sc.pl.umap(self.merged_adata,
                           color='celltype',
                           frameon=False,
                           wspace=0.6,
                           show=show,
                           palette=colors,
                           save=f'{dir_path}_celltypes.png' if dir_path else None)
            elif 'batch' in color_key:
                sc.pl.umap(self.merged_adata,
                           color='batch',
                           frameon=False,
                           wspace=0.6,
                           show=show,
                           save=f'{dir_path}_batch.png' if dir_path else None)

    def sankey_diagram(self, data, save_path=None, show=False, **kwargs):
        """Draws Sankey diagram for the given ``data``.
            Parameters
            ----------
            data: :class:`~numpy.ndarray`
                array with 2 columns. One for predictions and another for true values.
            save_path: str
                Path to save the drawn Sankey diagram. if ``None``, the diagram will not be saved.
            show: bool
                if ``True`` will show the diagram.
            kwargs:
                additional arguments for diagram configuration. See ``_alluvial.plot`` function.
        """
        font = {'family': 'Arial',
                'weight': 'bold',
                'size': 14}

        matplotlib.rc('font', **font)
        matplotlib.rc('ytick', labelsize=14)
        matplotlib.rc('xtick', labelsize=14)
        plt.close('all')
        ax = plot_alluvial(data.tolist(),
                            color_side=kwargs.get("color_side", 1),
                            alpha=kwargs.get("alpha", 0.5),
                            x_range=kwargs.get("x_range", (0, 1)),
                            res=kwargs.get("res", 20),
                            figsize=kwargs.get("figsize", (21, 15)),
                            disp_width=kwargs.get("disp_width", True),
                            width_in=kwargs.get("width_in", True),
                            wdisp_sep=kwargs.get("wdisp_sep", ' ' * 2),
                            cmap=kwargs.get("cmap", matplotlib.cm.get_cmap('jet')),
                            v_gap_frac=kwargs.get("v_gap_frac", 0.03),
                            h_gap_frac=kwargs.get("h_gap_frac", 0.03),
                            labels=kwargs.get("labels", None),
                            fontname=kwargs.get("fontname", "Arial"),
                            )
        if save_path is not None:
            plt.savefig(save_path, dpi=kwargs.get("dpi", 200), bbox_inches='tight')
        if show:
            plt.show()
        plt.close()




"""Garfield: Graph-based Contrastive Learning enable Fast Single-Cell Embedding"""
from ._settings import settings
from ._version import __version__

# import sys
# # needed when building doc (borrowed from scanpy)
# sys.modules.update(
#     {f'{__name__}.{m}': globals()[m] for m in ['pp']}) # ['tl', 'pp', 'pl']
    



"""
Functions for graph construction
"""
import numpy as np
import warnings
import scanpy as sc
import scipy.sparse as sp
import igraph as ig
import leidenalg
import pynndescent

from . import _utils as utils

def construct_graph_rna(
        adata, n_neighbors=15, n_pcs=None, metric='correlation', verbose=False
):
    """
    Compute k-nearest neighbors of data and return the UMAP graph.

    Parameters
    ----------
    adata: adata type of scanpy (n_samples, n_features)
        adata Data 
    n_neighbors: int
        Number of neighbors desired
    n_pcs: int
        Number PCs of executing graph construction
    metric: string, default='correlation'
        Metric used when constructing the initial knn graph
    verbose: bool, default=True
        Whether to print progress

    Returns
    -------
    rows, cols, vals: list
        Each edge is rows[i], cols[i], and its weight is vals[i]
    """

    if verbose:
        print("Constructing the graph...", flush=True)
    
    if sp.issparse(adata.X):
        adata.X = adata.X.toarray()
        
    # use scanpy functions to do the graph construction
    sc.pp.neighbors(adata, n_neighbors=n_neighbors, n_pcs=n_pcs, metric=metric)
    rows, cols = adata.obsp['connectivities'].nonzero()
    vals = adata.obsp['connectivities'][(rows, cols)].A1
    if verbose:
        print("Done!", flush=True)
    return rows, cols, vals

def leiden_clustering(n, edges, resolution=1, n_runs=1, seed=None, verbose=False):
    """
    Apply Leiden modularity maximization algorithm on the graph defined by edges and weights

    Parameters
    ----------
    n: int
        Number of edges in total
    edges: list of length two or three
        If length is two, then the graph is unweighted and each edge is (edges[0][i], edges[1][i]),
        if the length is three, then the graph is weighted and the weight of (edges[0][i], edges[1][i]) is edges[2][i].
    resolution: float, default=1
        Resolution parameter in Leiden algorithm
    n_runs: int, default=1
        Number of runs of Leiden algorithm, the run with the best modularity is taken as the output
    seed: None or int, default=None
        Random seed used. If None, use a random integer. If n_runs > 1, then seed will be reset to None.
    verbose: bool, default=True
        Whether to print progress

    Returns
    -------
    labels: np.array of shape (n,)
        Cluster labels
    """
    g = ig.Graph(directed=True)
    g.add_vertices(n)
    g.add_edges(list(zip(edges[0], edges[1])))
    if len(edges) > 2:
        g.es['weight'] = tuple(edges[2])

    if n_runs > 1 and seed is not None:
        seed = None
        warnings.warn('n_runs > 1, seed is reset to None.')

    partition_kwargs = {'n_iterations': -1, 'seed': seed,
                        'resolution_parameter': resolution}
    if len(edges) > 2:
        partition_kwargs['weights'] = np.array(g.es['weight']).astype(np.float64)

    partition_type = leidenalg.RBConfigurationVertexPartition

    best_modularity = float('-inf')
    best_labels = None
    if verbose and n_runs > 1:
        print('Running Leiden algorithm for {} times...'.format(n_runs), flush=True)
    for _ in range(n_runs):
        curr_part = leidenalg.find_partition(
            graph=g, partition_type=partition_type,
            **partition_kwargs
        )
        curr_modularity = curr_part.modularity
        if curr_modularity > best_modularity:
            best_modularity, best_labels = curr_modularity, np.array(curr_part.membership)

    assert best_labels is not None

    if verbose:
        if n_runs > 1:
            print('Best modularity among {} runs is {}.'.format(n_runs, best_modularity), flush=True)
        else:
            print('Modularity is {}.'.format(best_modularity), flush=True)
        print('The label has {} distinct clusters.'.format(len(np.unique(best_labels))), flush=True)

    return best_labels

def graph_clustering(
        n, edges, resolution=1, n_clusters=None, n_runs=1, resolution_tol=0.05, seed=None, verbose=False
):
    """
    Cluster the graph defined by edges and weights using Leiden algorithm.

    Parameters
    ----------
    n: int
        Number of edges in total
    edges: list of length two or three
        If length is two, then the graph is unweighted and each edge is (edges[0][i], edges[1][i]),
        if the length is three, then the graph is weighted and the weight of (edges[0][i], edges[1][i]) is edges[2][i].
    resolution: None or int, default=1
        If not None, then this is the resolution parameter in the clustering algorithm,
        if None, then n_clusters must be not None.
    n_clusters: None or int, default=None
        If not None, use binary search to select the resolution parameter to achieve the desired number of clusters,
        if None, then resolution must be not None.
    n_runs: int, default=1
        Number of runs of Leiden algorithm, the run with the best modularity is taken as the output.
    resolution_tol: float, default=0.05
        Any resolution within the range of plus/minus resolution_tol will not be differentiated.
    seed: None or int, default=None
        Random seed used. If None, use a random integer. If n_runs > 1, then seed will be reset to None.
    verbose: bool, default=True
        Whether to print progress

    Returns
    -------
    labels: np.array of shape (n,)
        Cluster labels
    """
    assert (resolution is not None) ^ (n_clusters is not None)

    def cluster_func(res, vb):
        return leiden_clustering(
            n=n, edges=edges, resolution=res, n_runs=n_runs, seed=seed, verbose=vb
        )

    if resolution is not None:
        return cluster_func(resolution, verbose)

    right = 1
    while True:
        curr_labels = cluster_func(right, False)
        curr_n_clusters = len(np.unique(curr_labels))
        if verbose:
            print('A resolution of {} gives {} clusters.'.format(right, curr_n_clusters), flush=True)
        if curr_n_clusters == n_clusters:
            return curr_labels
        elif curr_n_clusters < n_clusters:
            right *= 2
        else:
            break
    left = 0 if right == 1 else right / 2
    # desired resolution is in [left, right)
    while left + resolution_tol < right:
        mid = (left + right) / 2
        curr_labels = cluster_func(mid, False)
        curr_n_clusters = len(np.unique(curr_labels))
        if verbose:
            print('A resolution of {} gives {} clusters.'.format(mid, curr_n_clusters), flush=True)
        if curr_n_clusters == n_clusters:
            return curr_labels
        elif curr_n_clusters > n_clusters:
            right = mid
        else:
            left = mid
    return curr_labels


def get_nearest_neighbors(query_arr, target_arr, svd_components=None, randomized_svd=False, svd_runs=1,
                          metric='correlation'):
    """
    For each row in query_arr, compute its nearest neighbor in target_arr.

    Parameters
    ----------
    query_arr: np.array of shape (n_samples1, n_features)
        The query data matrix.
    target_arr: np.array of shape (n_samples2, n_features)
        The target data matrix.
    svd_components: None or int, default=None
        If not None, will first conduct SVD to reduce the dimension
        of the vertically stacked version of query_arr and target_arr.
    randomized_svd: bool, default=False
        Whether to use randomized SVD.
    svd_runs: int, default=1
        Run multiple instances of SVD and select the one with the lowest Frobenious reconstruction error.
    metric: string, default='correlation'
        The metric to use in nearest neighbor search.

    Returns
    -------
    neighbors: np.array of shape (n_samples1)
        The i-th element is the index in target_arr to whom the i-th row of query_arr is closest to.
    dists: np.array of shape (n_samples1)
        The i-th element is the distance corresponding to neighbors[i].
    """
    query_arr = utils.convert_to_numpy(query_arr)
    target_arr = utils.convert_to_numpy(target_arr)
    arr = np.vstack([query_arr, target_arr])
    arr = utils.svd_embedding(
        arr=arr, n_components=svd_components,
        randomized=randomized_svd,
        n_runs=svd_runs
    )
    query_arr = arr[:query_arr.shape[0], :]
    pivot_arr = arr[query_arr.shape[0]:, :]
    # approximate nearest neighbor search
    index = pynndescent.NNDescent(pivot_arr, n_neighbors=100, metric=metric)
    neighbors, dists = index.query(query_arr, k=50)
    neighbors, dists = neighbors[:, 0], dists[:, 0]
    return neighbors, dists



import numpy as np
import sys
import torch

class EarlyStopping:
    """
    Early stops the training if loss doesn't improve after a given patience.
    """
    def __init__(self, patience=10, verbose=False, checkpoint_file=''):
        """
        Parameters
        ----------
        patience
            How long to wait after last time loss improved. Default: 10
        verbose
            If True, prints a message for each loss improvement. Default: False
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.loss_min = np.Inf
        self.checkpoint_file = checkpoint_file

    def __call__(self, loss, model):
        if np.isnan(loss):
            self.early_stop = True
        score = -loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(loss, model)
        elif score <= self.best_score:
            self.counter += 1
            if self.verbose:
                print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
                model.load_model(self.checkpoint_file)
        else:
            self.best_score = score
            self.save_checkpoint(loss, model)
            self.counter = 0

    def save_checkpoint(self, loss, model):
        '''
        Saves model when loss decrease.
        '''
        if self.verbose:
            print(f'Loss decreased ({self.loss_min:.6f} --> {loss:.6f}).  Saving model ...')
        if self.checkpoint_file:
            torch.save(model.state_dict(), self.checkpoint_file)
        self.loss_min = loss


## progress function
def print_progress(epoch, logs, n_epochs=10000, only_val_losses=True):
    """Creates Message for '_print_progress_bar'.

       Parameters
       ----------
       epoch: Integer
            Current epoch iteration.
       logs: Dict
            Dictionary of all current losses.
       n_epochs: Integer
            Maximum value of epochs.
       only_val_losses: Boolean
            If 'True' only the validation dataset losses are displayed, if 'False' additionally the training dataset
            losses are displayed.

       Returns
       -------
    """
    message = ""
    for key in logs:
        if only_val_losses:
            if "val_" in key:
                message += f" - {key:s}: {logs[key][-1]:7.10f}"
        else:
            message += f" - {key:s}: {logs[key][-1]:7.10f}"

    _print_progress_bar(epoch + 1, n_epochs, prefix='', suffix=message, decimals=1, length=20)

def _print_progress_bar(iteration, total, prefix='', suffix='', decimals=1, length=100, fill='█'):
    """Prints out message with a progress bar.

       Parameters
       ----------
       iteration: Integer
            Current epoch.
       total: Integer
            Maximum value of epochs.
       prefix: String
            String before the progress bar.
       suffix: String
            String after the progress bar.
       decimals: Integer
            Digits after comma for all the losses.
       length: Integer
            Length of the progress bar.
       fill: String
            Symbol for filling the bar.

       Returns
       -------
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filled_len = int(length * iteration // total)
    bar = fill * filled_len + '-' * (length - filled_len)
    sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, bar, percent, '%', suffix)),
    if iteration == total:
        sys.stdout.write('\n')
    sys.stdout.flush()



"""Preprocessing"""
from ._utils import (
    get_centroids,
    summarize_clustering,
    drop_zero_variability_columns,
    robust_svd,
    svd_embedding,
    tfidf,
    GeneScores,
    gene_scores
)
from ._qc import (
    cal_qc_rna
)
from ._pca import (
    locate_elbow,
    select_pcs
)
from ._graph import (
    construct_graph_rna,
    leiden_clustering,
    graph_clustering,
    get_nearest_neighbors
)
from .read_adata import (
	read_mtx,
    read_scData,
    read_multi_scData,
    concat_data
)
from .preprocess import (
    preprocessing_rna,
    preprocessing_atac,
    preprocessing_adt,
    preprocessing # TODO
)


import os
import os.path as osp
import numpy as np
import pandas as pd

import torch
from torch_geometric.data import InMemoryDataset
from torch_geometric.data import Data
from sklearn.preprocessing import LabelEncoder

import scanpy as sc
import muon as mu
from mudata import MuData

# read data
from ..preprocessing.read_adata import concat_data
# preprocess utils
from ..preprocessing._utils import get_centroids, summarize_clustering
# QC
from ..preprocessing._qc import cal_qc_rna
# pca
from ..preprocessing._pca import  select_pcs
# graph
from ..preprocessing._graph import construct_graph_rna, graph_clustering
# preprocessing
from ..preprocessing.preprocess import preprocessing
# model_pre utils
from ._utils import Transfer_scData


class UserDataset(InMemoryDataset):
    def __init__(self, root, project_name, adata_list, profile, projection=False, data_type=None, sub_data_type=None,
                 weight=None, graph_const_method=None, genome=None, sample_col='batch', filter_cells_rna=False, min_features=100, min_cells=3,
                 keep_mt=False, use_top_pcs=False, use_gene_weigt=True, normalize=True, target_sum=1e4, used_hvg=True,
                 used_scale=True, single_n_top_genes=2000, rna_n_top_features=3000, atac_n_top_features=10000,
                 metacell_size=2, n_pcs=20, n_neighbors=15, metacell=True, metric='correlation',
                 method='umap', svd_solver='arpack', resolution_tol=0.1, leiden_runs=1,
                 leiden_seed=None, verbose=True):
        """
        UserDataset class for managing and processing single-cell or multi-modal datasets.

        Parameters:
        -----------
        root : str
            The directory where the graph dataset should be stored.
        project_name : str
            Name of the project for the dataset.
        adata_list : list
            List of AnnData objects representing different batches of data.
        profile : str
            Type of data profile, e.g., 'RNA', 'ATAC', 'multi-modal'.
        projection : bool
            Use for new dataset projection. Input the folder containing the pre-trained model. If False, don't do projection. Default: False.
        data_type : str, optional
            Type of multi-modal data, e.g., 'Paired', 'UnPaired'. Default is None.
        genome : str, optional
            Genome reference name, e.g., 'hg19', 'hg38', 'mm9', 'mm10'. Default is None, but if data_type is set as 'UnPaired', genome must be set.
        sample_col : str, optional
            Column name that defines different samples. Default is 'batch'.
        filter_cells_rna : bool, optional
            Whether to filter cells based on RNA expression. Default is False.
        min_features : int, optional
            Minimum number of features required for a cell. Default is 100.
        min_cells : int, optional
            Minimum number of cells required for a feature. Default is 3.
        keep_mt : bool, optional
            Whether to keep mitochondrial genes. Default is False.
        use_top_pcs : bool, optional
            Whether to use top principal components (PCs). Default is True.
        use_gene_weigt : bool, optional
            Whether to use gene weighting. Default is True.
        normalize : bool, optional
            Whether to normalize the data. Default is True.
        target_sum : float, optional
            Target sum for total count normalization. Default is 1e4.
        used_hvg : bool, optional
            Whether to use highly variable genes. Default is True.
        used_scale : bool, optional
            Whether to scale the data. Default is True.
        single_n_top_genes : int, optional
            Number of top genes to select for single-sample analysis. Default is 2000.
        rna_n_top_features : int, optional
            Number of top features to select for RNA profile. Default is 3000.
        atac_n_top_features : int, optional
            Number of top features to select for ATAC profile. Default is 10000.
        metacell_size : int, optional
            Size of metacell clusters, which will determine the number of metacell clusters. Default is 2.
        n_pcs : int, optional
            Number of principal components to use for each sample processing. Default is 20.
        n_neighbors : int, optional
            Number of neighbors for graph construction. Default is 15.
        metacell : bool, optional
            Whether to construct metacells. Default is True.
        metric : str, optional
            Distance metric for nearest neighbor search. Default is 'correlation'.
        method : str, optional
            Dimensionality reduction method. Default is 'umap'.
        svd_solver : str, optional
            SVD solver to use for PCA. Default is 'arpack'.
        resolution_tol : float, optional
            Tolerance level for resolution clustering. Default is 0.1.
        leiden_runs : int, optional
            Number of times to run Leiden clustering. Default is 1.
        leiden_seed : int or None, optional
            Seed for the Leiden algorithm. Default is None.
        verbose : bool, optional
            Whether to print progress information. Default is True.
        """

        self.name = project_name
        self.data_list = adata_list
        self.profile = profile
        self.projection = projection
        self.data_type = data_type
        self.sub_data_type = sub_data_type
        self.weight = weight
        self.graph_const_method = graph_const_method
        self.genome = genome
        self.use_gene_weigt = use_gene_weigt
        self.use_top_pcs = use_top_pcs
        self.sample_col = sample_col
        self.filter_cells_rna = filter_cells_rna
        self.min_features = min_features
        self.min_cells = min_cells
        self.keep_mt = keep_mt
        self.normalize = normalize
        self.target_sum = target_sum
        self.used_hvg = used_hvg
        self.used_scale = used_scale
        self.single_n_top_genes = single_n_top_genes
        self.rna_n_top_features = rna_n_top_features
        self.atac_n_top_features = atac_n_top_features
        self.metacell_size = metacell_size
        self.metacell = metacell
        self.n_pcs = n_pcs
        self.n_neighbors = n_neighbors
        self.metric = metric
        self.method = method
        self.svd_solver = svd_solver
        self.resolution_tol = resolution_tol
        self.leiden_runs = leiden_runs
        self.leiden_seed = leiden_seed
        self.verbose = verbose

        super(UserDataset, self).__init__(root)
        try:
            print("Data files exist, loading...")
            self.data = torch.load(self.processed_paths[0])
            ## 加载single-cell or metacell的merged_adata对象
            path = osp.join(self.processed_dir, f'{self.name}_merged_adata.pt')
            self.merged_adata = torch.load(path)
            print("Finished loading.")
        except FileNotFoundError:
            print("Data files do not exist, generating...")
            self.process()
            self.data = torch.load(self.processed_paths[0])
            ## 加载single-cell or metacell的merged_adata对象
            path = osp.join(self.processed_dir, f'{self.name}_merged_adata.pt')
            self.merged_adata = torch.load(path)
            print("Finished generating.")

    @property
    def processed_dir(self) -> str:
        return osp.join(self.root, self.name + '_processed')

    # 检查self.processed_dir目录下是否存在self.processed_file_names属性方法返回的所有文件，没有就会走process
    @property
    def processed_file_names(self):
        return ['User_single_cell.dataset']

    def process_sample(self, adata):
        # 创建一个包含子对象的列表
        sub_adata_list = []

        # 样本划分：假设每个样本的样本名称存储在adata.obs['sample_name']中
        sample_names = np.unique(adata.obs[self.sample_col])

        for sample_name in sample_names:
            # 选择属于当前样本的观测
            sub_adata = adata[adata.obs[self.sample_col] == sample_name].copy()

            # Filter genes for RNA dataset...
            sc.pp.filter_genes(sub_adata, min_cells=self.min_cells)

            # Perform QC for RNA dataset...
            cal_qc_rna(sub_adata, expr_cutoff=1)

            # Filter out cells if needed:
            if self.filter_cells_rna:
                sc.pp.filter_cells(sub_adata, min_genes=self.min_features)

            # normalization
            if self.normalize:
                # 校正测序深度
                sc.pp.normalize_total(sub_adata, target_sum=self.target_sum)
                # 表达值log转化
                sc.pp.log1p(sub_adata)
                # 备份完整矩阵
                sub_adata.raw = sub_adata

            # Optionally, variable gene selection step can be also performed.
            if self.used_hvg:
                sc.pp.highly_variable_genes(sub_adata, n_top_genes=self.single_n_top_genes)

            if self.used_scale:
                # scale data, clip values exceeding standard deviation 10.
                sc.pp.scale(sub_adata, max_value=10)

            # 将子对象添加到列表中
            sub_adata_list.append(sub_adata)

        return sub_adata_list

    def _construct_graphs(self, sub_adata_list=None, metacell_labels=None, used_hvg=True,
                          n_top_genes=2000, n_pcs=20, n_neighbors=15, metacell=False,
                          randomized_svd=False, svd_runs=1, metric='correlation',
                          svd_solver='arpack', verbose=True):
        """
        Construct neighborhood graphs for all batches of the target array.
        """
        # 创建一个包含metacell子对象的列表
        if metacell_labels is not None:
            self._metacell_labels1 = metacell_labels
        self.meta_adata_sub_list = []
        meta_adata_sub_list = self.meta_adata_sub_list
        self._edges1 = []
        edges = self._edges1

        for b, sub_adata in enumerate(sub_adata_list):
            if verbose:
                print('Constructing neighborhood graphs for cells in sample{}...'.format(b), flush=True)

            ## meta cell 构建，在单细胞的图结构的基础上
            if metacell_labels is not None and metacell:
                meta_adata_sub = get_centroids(sub_data=sub_adata,
                                               labels=self._metacell_labels1[b])

                ## 将counts矩阵赋值给adata.X,方便重新标准化
                if meta_adata_sub.layers['counts'] is not None:
                    meta_adata_sub.X = meta_adata_sub.layers['counts'].copy()

                # 将子对象添加到列表中
                meta_adata_sub_list.append(meta_adata_sub)
            else:
                ## 单细胞构图
                # 数据PCA降维，为构建图准备数据
                if used_hvg:
                    sc.tl.pca(sub_adata, svd_solver=svd_solver, use_highly_variable=used_hvg)

                # 自动确定 pc 的数目
                print('Auto-query the number of PCs for single-cell-graph construction of sample{}...'.format(b),
                      flush=True)
                if n_pcs is not None:
                    select_pcs(sub_adata, n_pcs=n_pcs)
                else:
                    select_pcs(sub_adata, n_pcs=None)
                n_pcs = sub_adata.uns['pca']['n_pcs']
                print('The best n_pcs is {}'.format(n_pcs), flush=True)

                edges.append(
                    construct_graph_rna(
                        adata=sub_adata,
                        n_pcs=n_pcs,
                        n_neighbors=n_neighbors,
                        metric=metric,
                        verbose=False
                    )
                )

        if verbose:
            print('Graph construction finished!', flush=True)
        return meta_adata_sub_list, edges

    def _cluster_graphs(self, sub_adata_list=None, metacell_labels=None, used_hvg=True,
                        n_top_genes=2000, n_pcs=20, n_neighbors=15, metacell_size=2,
                        metacell=False, randomized_svd=False, svd_runs=1, metric='correlation',
                        resolution_tol=0.1, leiden_runs=1, leiden_seed=None, svd_solver='arpack',
                        verbose=True):
        """
        Cluster the neighborhood graphs.
        """
        if not metacell:
            ## single cell clustering
            _, edges = self._construct_graphs(
                sub_adata_list=sub_adata_list, metacell_labels=metacell_labels,
                n_top_genes=n_top_genes, used_hvg=used_hvg,
                n_pcs=n_pcs, n_neighbors=n_neighbors,
                metacell=metacell, randomized_svd=randomized_svd, svd_runs=svd_runs,
                metric=metric, svd_solver=svd_solver, verbose=verbose
            )  ## 需要执行两次，第一次执行单细胞合并时，adata_sub_list是空的，

            self.metacell_size = metacell_size
            if self.metacell_size > 1:
                if verbose:
                    print(
                        'Aggregating cells into metacells of average size {}...'.format(self.metacell_size),
                        flush=True
                    )
                print('Single-Cell Merging...', flush=True)
                self._metacell_labels1 = []
                metacell_labels = self._metacell_labels1
                for b, curr_edges in enumerate(edges):
                    if verbose:
                        print('Now at sample {}...'.format(b), flush=True)
                    n = len(sub_adata_list[b])
                    metacell_labels.append(
                        graph_clustering(
                            n=n,
                            edges=curr_edges,
                            n_clusters=int(n // self.metacell_size),
                            resolution=None,
                            n_runs=leiden_runs,
                            resolution_tol=resolution_tol,
                            seed=leiden_seed,
                            verbose=False
                        )
                    )
                if verbose:
                    print('Single-cell merging finished!', flush=True)
                return metacell_labels
        else:
            if verbose:
                print(
                    'Constructing meta-cells scRNA object',
                    flush=True
                )
            meta_adata_sub_list, _ = self._construct_graphs(
                sub_adata_list=sub_adata_list, metacell_labels=metacell_labels,
                n_top_genes=n_top_genes, used_hvg=used_hvg,
                n_pcs=n_pcs, n_neighbors=n_neighbors,
                metacell=metacell, randomized_svd=randomized_svd, svd_runs=svd_runs,
                metric=metric, svd_solver=svd_solver, verbose=verbose
            )  ## 第二次执行后，里面存了合并metacell的单细胞对象

            return meta_adata_sub_list  # , labels

    def construct_metacell(self, sub_adata_list, used_hvg=True, n_top_genes=2000, n_pcs=20,
                         n_neighbors=15, metacell_size=2, metacell=True, randomized_svd=False, svd_runs=1,
                         metric='correlation', resolution_tol=0.1, leiden_runs=1,
                         leiden_seed=None, svd_solver='arpack', verbose=True):
        """
        Construct neighborhood graphs and cluster them as needed.

        Parameters
        ----------
        n_neighbors: int, default=15
            Number of neighbors for graph construction for arr.
        randomized_svd: bool, default=False
            Whether to perform randomized SVD.
        svd_runs: int, default=1
            Perform multiple runs of SVD and select the one with lowest Frobenious reconstruction error is selected.
        resolution_tol: float, default=0.1
            Any resolution within the range of plus/minus resolution_tol will not be differentiated.
        leiden_runs: int, default=1
            Perform multiple runs of Leiden algorithm and the one with highest modularity is selected.
        metric: string, default='correlation'
            The metric to use in nearest neighbor search.
        leiden_seed: None or int, default=None
            Random seed for Leiden algorithm. If leiden_runs>1, leiden_seed will be reset to None.
        verbose: bool, default=True
            Whether to print the progress.

        Returns
        -------
        None
        """
        self.metacell_size = metacell_size
        if self.metacell_size > 1:
            # single cell clustering
            # need to construct the graph + metacell merging
            metacell_labels = self._cluster_graphs(sub_adata_list=sub_adata_list, metacell_labels=None,
                                                   n_top_genes=n_top_genes, used_hvg=used_hvg,
                                                   n_pcs=n_pcs, n_neighbors=n_neighbors,
                                                   metacell_size=metacell_size,
                                                   metacell=False, randomized_svd=randomized_svd,
                                                   svd_runs=svd_runs, metric=metric,
                                                   resolution_tol=resolution_tol,
                                                   leiden_runs=leiden_runs, leiden_seed=leiden_seed,
                                                   svd_solver=svd_solver, verbose=verbose
                                                   )
            # meta cell scRNA constrution
            meta_adata_sub_list = self._cluster_graphs(
                sub_adata_list=sub_adata_list, metacell_labels=metacell_labels, used_hvg=used_hvg,
                n_top_genes=n_top_genes, n_pcs=n_pcs, n_neighbors=n_neighbors,
                metacell_size=metacell_size, metacell=metacell, randomized_svd=randomized_svd,
                svd_runs=svd_runs, metric=metric, resolution_tol=resolution_tol,
                leiden_runs=leiden_runs, leiden_seed=leiden_seed, svd_solver=svd_solver,
                verbose=verbose
            )

        return meta_adata_sub_list

    def process(self):
        # load data
        adata = concat_data(
            self.data_list,
            batch_categories=None,
            join='inner',
            batch_key=self.sample_col, # 'batch'
            index_unique=None,
            save=None
        )
        if isinstance(adata, sc.AnnData):
            if adata.X.max() < 50:
                print('Warning: adata.X may have already been normalized, adata.X must be `counts`, please check.')
            else:
                adata.layers['counts'] = adata.X.copy()
        elif isinstance(adata, mu.MuData):
            if adata.mod['rna'].X.max() < 50:
                print('Warning: adata.X may have already been normalized, adata.X must be `counts`, please check.')
            else:
                adata.mod['rna'].layers['counts'] = adata.mod['rna'].X.copy()

        # RNA ATAC ADT
        if self.profile in ['RNA', 'ATAC', 'ADT']:
            rna_adata = adata.copy()
            del adata

            if self.metacell and self.metacell_size > 1:
                # 根据样本为单位进行预处理
                sub_adata_list = self.process_sample(rna_adata)
                # 进行图构建+聚类，返回每个metacell的单细胞对象
                # 注意如果used_hvg为True，则只用高变异基因构建图
                metacell_adata_list = self.construct_metacell(
                    sub_adata_list, used_hvg=self.used_hvg, n_top_genes=self.single_n_top_genes, n_pcs=self.n_pcs,
                    n_neighbors=self.n_neighbors, metacell_size=self.metacell_size, metacell=self.metacell,
                    metric=self.metric, resolution_tol=self.resolution_tol, leiden_runs=self.leiden_runs,
                    leiden_seed=self.leiden_seed, svd_solver=self.svd_solver, verbose=self.verbose
                )
                rna_adata = metacell_adata_list[0].concatenate(metacell_adata_list[1:], join='inner')  ## X 的 数据是counts
            else:
                rna_adata = rna_adata.copy()

            ## 预处理
            _, rna_adata_hvg = preprocessing(
                rna_adata,
                profile=self.profile,
                data_type=self.data_type,
                genome=self.genome,
                use_gene_weigt=self.use_gene_weigt,
                use_top_pcs=self.use_top_pcs,
                min_features=self.min_features,
                min_cells=self.min_cells,
                target_sum=self.target_sum,
                rna_n_top_features=self.rna_n_top_features,
                atac_n_top_features=self.atac_n_top_features,
                n=self.n_neighbors,
                batch_key=self.sample_col,
                metric=self.metric,
                svd_solver=self.svd_solver,
                keep_mt=self.keep_mt
            )

            # 保存rna_adata结果，用于模型输入
            path = osp.join(self.processed_dir, f'{self.name}_merged_adata.pt')
            torch.save(rna_adata_hvg, path, pickle_protocol=4)

            # 构建 pytorch_geometric 的数据结构
            data = Transfer_scData(rna_adata_hvg, self.sample_col)

        ### Paired multi-modal
        elif self.profile == 'multi-modal':
            if len(self.sub_data_type) == 2:
                if self.sub_data_type[0] == 'rna':
                    rna_adata = adata.mod['rna'].copy()
                elif self.sub_data_type[1] == 'atac':
                    atac_adata = adata.mod['atac'].copy()
                elif self.sub_data_type[1] == 'adt':
                    adt_adata = adata.mod['adt'].copy()
            else:
                ValueError('The length of sub_data_type must be 2, such as: ["rna", "atac"] or ["rna", "adt"].')
            del adata

            if self.metacell and self.metacell_size > 1:
                ## 根据样本为单位进行预处理
                sub_adata_list = self.process_sample(rna_adata)
                ## 进行图构建+聚类，返回每个metacell的单细胞对象
                ## 注意如果used_hvg为True，则只用高变异基因构建图
                metacell_adata_list = self.construct_metacell(
                    sub_adata_list, used_hvg=self.used_hvg, n_top_genes=self.single_n_top_genes, n_pcs=self.n_pcs,
                    n_neighbors=self.n_neighbors, metacell_size=self.metacell_size, metacell=self.metacell,
                    metric=self.metric, resolution_tol=self.resolution_tol, leiden_runs=self.leiden_runs,
                    leiden_seed=self.leiden_seed, svd_solver=self.svd_solver, verbose=self.verbose
                )
                rna_adata = metacell_adata_list[0].concatenate(metacell_adata_list[1:], join='inner')  ## X 的 数据是counts
            else:
                rna_adata = rna_adata.copy()

            if atac_adata is not None:
                mdata = MuData({"rna": rna_adata, "atac": atac_adata})
            elif adt_adata is not None:
                mdata = MuData({"rna": rna_adata, "adt": adt_adata})

            ## 预处理
            merged_adata = preprocessing(
                mdata,
                profile=self.profile,
                data_type=self.data_type,
                sub_data_type=self.sub_data_type,
                genome=self.genome,
                weight=self.weight,
                use_gene_weigt=self.use_gene_weigt,
                use_top_pcs=self.use_top_pcs,
                min_features=self.min_features,
                min_cells=self.min_cells,
                target_sum=self.target_sum,
                rna_n_top_features=self.rna_n_top_features,
                atac_n_top_features=self.atac_n_top_features,
                n=self.n_neighbors,
                batch_key=self.sample_col,
                metric=self.metric,
                svd_solver=self.svd_solver,
                keep_mt=self.keep_mt
            )

            ## 保存merged_adata结果，用于模型输入
            path = osp.join(self.processed_dir, f'{self.name}_merged_adata.pt')
            torch.save(merged_adata, path, pickle_protocol=4)

            ## 构建 pytorch_geometric 的数据结构
            data = Transfer_scData(merged_adata, self.sample_col, self.data_type)

        ### spatial single- or multi-modal
        elif self.profile == 'spatial':
            ## 预处理
            merged_adata = preprocessing(
                adata,
                profile=self.profile,
                data_type=self.data_type,
                sub_data_type=self.sub_data_type,
                genome=self.genome,
                weight=self.weight,
                graph_const_method=self.graph_const_method,
                use_gene_weigt=self.use_gene_weigt,
                use_top_pcs=self.use_top_pcs,
                min_features=self.min_features,
                min_cells=self.min_cells,
                target_sum=self.target_sum,
                rna_n_top_features=self.rna_n_top_features,
                atac_n_top_features=self.atac_n_top_features,
                n=self.n_neighbors,
                batch_key=self.sample_col,
                metric=self.metric,
                svd_solver=self.svd_solver,
                keep_mt=self.keep_mt
            )
            ## 保存merged_adata结果，用于模型输入
            path = osp.join(self.processed_dir, f'{self.name}_merged_adata.pt')
            torch.save(merged_adata, path, pickle_protocol=4)

            ## 构建 pytorch_geometric 的数据结构
            data = Transfer_scData(merged_adata, self.sample_col, self.data_type)

        else:
            return "Unknown input data type."

        print('Data preprocessing and graph construction finished')
        torch.save(data, self.processed_paths[0])

    def __repr__(self) -> str:
        return f'{self.name}({len(self)})'


import inspect
import os
import dill
# import pickle
from copy import deepcopy
from typing import Optional, Union

import numpy as np
import torch
import scanpy as sc
from anndata import AnnData, concat
from scipy.sparse import issparse

from ._utils import UnpicklerCpu, _validate_var_names


class BaseMixin:
    """ Adapted from
        Title: scvi-tools
        Authors: Romain Lopez <romain_lopez@gmail.com>,
                 Adam Gayoso <adamgayoso@berkeley.edu>,
                 Galen Xing <gx2113@columbia.edu>
        Date: 14.12.2020
        Code version: 0.8.0-beta.0
        Availability: https://github.com/YosefLab/scvi-tools
        Link to the used code:
        https://github.com/YosefLab/scvi-tools/blob/0.8.0-beta.0/scvi/core/models/base.py
    """

    def _get_user_attributes(self):
        # returns all the self attributes defined in a model class, eg, self.is_trained_
        attributes = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        attributes = [
            a for a in attributes if not (a[0].startswith("__") and a[0].endswith("__"))
        ]
        attributes = [a for a in attributes if not a[0].startswith("_abc_")]
        return attributes

    def _get_public_attributes(self):
        public_attributes = self._get_user_attributes()
        public_attributes = {a[0]: a[1] for a in public_attributes if a[0][-1] == "_"}
        return public_attributes

    def save(
            self,
            dir_path: str,
            overwrite: bool = False,
            save_anndata: bool = True,
            **anndata_write_kwargs,
    ):
        """Save the state of the model.
           Neither the trainer optimizer state nor the trainer history are saved.
           Parameters
           ----------
           dir_path
                Path to a directory.
           overwrite
                Overwrite existing data or not. If `False` and directory
                already exists at `dir_path`, error will be raised.
           save_anndata
                If True, also saves the anndata
           anndata_write_kwargs
                Kwargs for anndata write function
        """
        # get all the public attributes
        public_attributes = self._get_public_attributes()
        # save the model state dict and the trainer state dict only
        if not os.path.exists(dir_path) or overwrite:
            os.makedirs(dir_path, exist_ok=overwrite)
        else:
            raise ValueError(
                "{} already exists. Please provide an unexisting directory for saving.".format(
                    dir_path
                )
            )

        if save_anndata:
            self.merged_adata.write(
                os.path.join(dir_path, "adata_ref.h5ad"), **anndata_write_kwargs
            )

        model_save_path = os.path.join(dir_path, "model_params.pt")
        attr_save_path = os.path.join(dir_path, "attr.pkl")
        varnames_save_path = os.path.join(dir_path, "var_names.csv")

        var_names = self.merged_adata.var_names.astype(str)
        var_names = var_names.to_numpy()
        np.savetxt(varnames_save_path, var_names, fmt="%s")

        torch.save(self.model.state_dict(), model_save_path)

        # 现在使用 pickle.dump 来序列化，pickle 库已被替换为 dill
        with open(attr_save_path, "wb") as f:
            dill.dump(public_attributes, f)

    def _load_expand_params_from_dict(self, state_dict):
        load_state_dict = state_dict.copy()

        new_state_dict = self.model.state_dict()

        # 创建一个新的状态字典，仅包括在 new_state_dict 中也存在的键
        load_state_dict = {k: v for k, v in load_state_dict.items() if k in new_state_dict}

        for key, ten in new_state_dict.items():
            if key not in load_state_dict:
                load_state_dict[key] = ten

        self.model.load_state_dict(load_state_dict)

    @classmethod
    def _load_params(cls, dir_path: str, map_location: Optional[str] = None):
        setup_dict_path = os.path.join(dir_path, "attr.pkl")
        model_path = os.path.join(dir_path, "model_params.pt")
        varnames_path = os.path.join(dir_path, "var_names.csv")

        try:
            with open(setup_dict_path, "rb") as handle:
                attr_dict = dill.load(handle)
        # This catches the following error:
        # RuntimeError: Attempting to deserialize object on a CUDA device
        # but torch.cuda.is_available() is False.
        # If you are running on a CPU-only machine, please use torch.load with
        # map_location=torch.device('cpu') to map your storages to the CPU.
        except RuntimeError:
            with open(setup_dict_path, "rb") as handle:
                attr_dict = UnpicklerCpu(handle).load()

        model_state_dict = torch.load(model_path, map_location=map_location)

        var_names = np.genfromtxt(varnames_path, delimiter=",", dtype=str)

        return attr_dict, model_state_dict, var_names


class SurgeryMixin:
    @classmethod
    def load_query_data(
        cls,
        adata: AnnData,
        reference_model: Union[str, 'Model'],
        freeze: bool = True,
        remove_dropout: bool = True,
        map_location = None,
        **kwargs
    ):
        """Transfer Learning function for new data. Uses old trained model and expands it for new conditions.

           Parameters
           ----------
           adata
                Query anndata object.
           reference_model
                A model to expand or a path to a model folder.
           freeze: Boolean
                If 'True' freezes every part of the network except the first layers of encoder/decoder.
           remove_dropout: Boolean
                If 'True' remove Dropout for Transfer Learning.
           map_location
                map_location to remap storage locations (as in '.load') of 'reference_model'.
                Only taken into account if 'reference_model' is a path to a model on disk.
           kwargs
                kwargs for the initialization of the query model.

           Returns
           -------
           new_model
                New model to train on query data.
        """
        if isinstance(reference_model, str):
            attr_dict, model_state_dict, var_names = cls._load_params(reference_model, map_location)
            adata_ref = sc.read_h5ad(os.path.join(reference_model, "adata_ref.h5ad"))
            adata_ref.X = adata_ref.layers["counts"] if "counts" in adata_ref.layers.keys() else adata_ref.X
            # adata = _validate_var_names(adata, var_names)
        else:
            attr_dict = reference_model._get_public_attributes()
            model_state_dict = reference_model.model.state_dict()
            var_names = reference_model.merged_adata.var_names
            adata_ref = reference_model.merged_adata
            adata_ref.X = adata_ref.layers["counts"] if "counts" in adata_ref.layers.keys() else adata_ref.X
            # adata = _validate_var_names(adata, reference_model.merged_adata.var_names)

        init_params = deepcopy(cls._get_init_params_from_dict(attr_dict))

        if remove_dropout:
            init_params['dropout'] = 0.0

        # don't preprocess the data
        init_params['min_features'] = 0
        init_params['min_cells'] = 0

        # load the data
        adata_concat = concat(
            [adata_ref, adata],
            label='projection',
            keys=['reference', 'query'],
            index_unique=None,
        )
        init_params['adata_list'] = adata_concat
        init_params['rna_n_top_features'] = var_names

        init_params.update(kwargs)

        new_model = cls(init_params)
        new_model.model.to(next(iter(model_state_dict.values())).device)
        new_model._load_expand_params_from_dict(model_state_dict)

        if freeze:
            new_model.model.freeze = True
            for name, p in new_model.model.named_parameters():
                p.requires_grad = False
                if 'eps_weight' in name or 'eps_bias' in name:
                    p.requires_grad = True
                if "layers.0" in name or "norm.0" in name:
                    p.requires_grad = True

        return new_model


import numpy as np
import scipy
from sklearn.neighbors import NearestNeighbors, KNeighborsRegressor
from sklearn.metrics import silhouette_samples, silhouette_score

def batch_entropy_mixing_score(data, batches, n_neighbors=100, n_pools=100, n_samples_per_pool=100):
    """
    Calculate batch entropy mixing score

    Algorithm
    ---------
        * 1. Calculate the regional mixing entropies at the location of 100 randomly chosen cells from all batches
        * 2. Define 100 nearest neighbors for each randomly chosen cell
        * 3. Calculate the mean mixing entropy as the mean of the regional entropies
        * 4. Repeat above procedure for 100 iterations with different randomly chosen cells.

    Parameters
    ----------
    data
        np.array of shape nsamples x nfeatures.
    batches
        batch labels of nsamples.
    n_neighbors
        The number of nearest neighbors for each randomly chosen cell. By default, n_neighbors=100.
    n_samples_per_pool
        The number of randomly chosen cells from all batches per iteration. By default, n_samples_per_pool=100.
    n_pools
        The number of iterations with different randomly chosen cells. By default, n_pools=100.

    Returns
    -------
    Batch entropy mixing score
    """

    #     print("Start calculating Entropy mixing score")
    def entropy(batches):
        p = np.zeros(N_batches)
        adapt_p = np.zeros(N_batches)
        a = 0
        for i in range(N_batches):
            p[i] = np.mean(batches == batches_[i])
            a = a + p[i] / P[i]
        entropy = 0
        for i in range(N_batches):
            adapt_p[i] = (p[i] / P[i]) / a
            entropy = entropy - adapt_p[i] * np.log(adapt_p[i] + 10 ** -8)
        return entropy

    n_neighbors = min(n_neighbors, len(data) - 1)
    nne = NearestNeighbors(n_neighbors=1 + n_neighbors, n_jobs=8)
    nne.fit(data)
    kmatrix = nne.kneighbors_graph(data) - scipy.sparse.identity(data.shape[0])

    score = 0
    batches_ = np.unique(batches)
    N_batches = len(batches_)
    if N_batches < 2:
        raise ValueError("Should be more than one cluster for batch mixing")
    P = np.zeros(N_batches)
    for i in range(N_batches):
        P[i] = np.mean(batches == batches_[i])
    for t in range(n_pools):
        indices = np.random.choice(np.arange(data.shape[0]), size=n_samples_per_pool)
        score += np.mean([entropy(batches[kmatrix[indices].nonzero()[1]
        [kmatrix[indices].nonzero()[0] == i]])
                          for i in range(n_samples_per_pool)])
    Score = score / float(n_pools)
    return Score / float(np.log2(N_batches))


def silhouette(
        X,
        cell_type,
        metric='euclidean',
        scale=True
):
    """
    Wrapper for sklearn silhouette function values range from [-1, 1] with
        1 being an ideal fit
        0 indicating overlapping clusters and
        -1 indicating misclassified cells
    By default, the score is scaled between 0 and 1. This is controlled `scale=True`

    :param group_key: key in adata.obs of cell labels
    :param embed: embedding key in adata.obsm, default: 'X_pca'
    :param scale: default True, scale between 0 (worst) and 1 (best)
    """
    asw = silhouette_score(
        X,
        cell_type,
        metric=metric
    )
    if scale:
        asw = (asw + 1) / 2
    return asw




import math
import numpy as np
import scipy.sparse as sp
from typing import Optional, Tuple

import torch
import torch.nn as nn
from torch import Tensor
import torch.nn.functional as F
from torch.nn.functional import normalize
from torch_geometric.utils import (negative_sampling, remove_self_loops, add_self_loops)
from torch_geometric.data import Data
from torch_geometric.nn import GCNConv, GATConv, GATv2Conv
from torch_geometric.nn import GAE, InnerProductDecoder

from ._utils import scipy_sparse_mat_to_torch_sparse_tensor, extract_subgraph

# Domain-specific Batch Normalization（领域特定的批次归一化）是一种调整神经网络中批次归一化层以适应不同输入数据源或领域的技术。
class DSBatchNorm(nn.Module):
    """
    Domain-specific Batch Normalization
    """

    def __init__(self, num_features, n_domain, eps=1e-5, momentum=0.1):
        """
        Parameters
        ----------
        num_features
            dimension of the features
        n_domain
            domain number
        """
        super().__init__()
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        # self.device = torch.device("cpu")
        self.n_domain = n_domain
        self.num_features = num_features
        self.bns = nn.ModuleList(
            [nn.BatchNorm1d(num_features, eps=eps, momentum=momentum) for i in range(n_domain)]).to(self.device)

    def reset_running_stats(self):
        for bn in self.bns:
            bn.reset_running_stats()

    def reset_parameters(self):
        for bn in self.bns:
            bn.reset_parameters()

    def _check_input_dim(self, input):
        raise NotImplementedError

    def forward(self, x, y):
        device = x.device  # 获取输入数据的设备
        out = torch.zeros(x.size(0), self.num_features, device=device)  # , requires_grad=False)
        for i in range(self.n_domain):
            indices = np.where(y.cpu().numpy() == i)[0]
            if len(indices) > 1:
                out[indices] = self.bns[i](x[indices])
            elif len(indices) == 1:
                self.bns[i].training = False
                out[indices] = self.bns[i](x[indices])
                self.bns[i].training = True
        return out


class GATEncoder(nn.Module):
    def __init__(self, in_channels, hidden_dims, latent_dim, conv_type, svd_q, num_heads,
                 dropout, concat, num_domains='', used_edge_weight=False, used_DSBN=False):
        """
        Initializes the GATEncoder

        Parameters
        ----------
        in_channels : int
            Number of input feature dimensions (length of each node's feature vector).
        hidden_dims : list[int]
            List of output dimensions for each hidden layer in the GAT.
        latent_dim : int
            Dimension of the latent feature representation produced by the encoder.
        svd_q : int
            Rank for the low-rank SVD approximation. Default is 5.
        num_heads : int
            Number of attention heads for each GAT layer.
        dropout : float
            Dropout rate for GAT layers.
        concat : bool
            Whether to concatenate outputs of all attention heads or not.
        num_domains : int or str
            Number of domains for domain-specific batch normalization (DSBN). If `1`, regular batch normalization is used.
        used_edge_weight : bool, optional
            Whether to use edge weights in the GAT layers (default is False).
        used_DSBN : bool, optional
            Whether to use domain-specific batch normalization (DSBN) (default is False).
        """
        super(GATEncoder, self).__init__()
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        # self.device = torch.device("cpu")
        self.svd_q = svd_q
        self.used_DSBN = used_DSBN
        self.used_edge_weight = used_edge_weight
        self.layers = nn.ModuleList()
        self.norm = nn.ModuleList()
        if conv_type == 'GAT':
            GATLayer = GATConv
        elif conv_type == 'GATv2Conv':
            GATLayer = GATv2Conv

        num_hidden_layers = len(hidden_dims)
        num_heads_list = [num_heads] * num_hidden_layers
        concat_list = [concat] * num_hidden_layers
        for i in range(num_hidden_layers):
            if concat_list[i]:
                current_dim = hidden_dims[i] * num_heads_list[i]
            else:
                current_dim = hidden_dims[i]

            if type(num_domains) == int:
                if num_domains == 1:  # TO DO
                    norm = nn.BatchNorm1d(current_dim)
                else:
                    # num_domains >1 represent domain-specific batch normalization of n domain
                    norm = DSBatchNorm(current_dim, num_domains)
            else:
                norm = None
            self.norm.append(norm)

        current_dim = in_channels
        # num_heads_list = [num_heads] * num_hidden_layers
        dropout_list = [dropout] * num_hidden_layers
        for i in range(num_hidden_layers):
            layer = GATLayer(
                in_channels=current_dim,
                out_channels=hidden_dims[i],
                heads=num_heads_list[i],
                dropout=dropout_list[i],
                concat=concat_list[i],
                edge_dim=1 if self.used_edge_weight else None
            )
            self.layers.append(layer)
            if concat_list[i]:
                current_dim = hidden_dims[i] * num_heads_list[i]
            else:
                current_dim = hidden_dims[i]

        self.conv_mean = GATLayer(in_channels=current_dim,
                                  out_channels=latent_dim,
                                  heads=num_heads,
                                  concat=False,
                                  edge_dim=1 if self.used_edge_weight else None,
                                  dropout=dropout)
        self.conv_log_std = GATLayer(in_channels=current_dim,
                                     out_channels=latent_dim,
                                     heads=num_heads,
                                     concat=False,
                                     edge_dim=1 if self.used_edge_weight else None,
                                     dropout=dropout)

    def forward(self, data):
        """
        Basic block consist of:
            fc -> bn -> act -> dropout
        """
        x1, edge_index1, y1, edge_weight1 = data.x, data.edge_index, data.y, data.edge_attr

        # 首先，将edge_index和edge_attr转换为稀疏张量
        num_nodes = data.num_nodes  # 假设数据中节点的总数
        # 创建 SciPy 的 COO 格式稀疏矩阵
        value = torch.squeeze(data.edge_attr).cpu().numpy()
        row = data.edge_index[0].cpu().numpy()
        col = data.edge_index[1].cpu().numpy()
        sparse_adj = sp.coo_matrix((value, (row, col)),
                                   shape=(num_nodes, num_nodes))
        # perform svd reconstruction
        adj = scipy_sparse_mat_to_torch_sparse_tensor(sparse_adj).coalesce()
        #         print('Performing SVD...')
        svd_u, s, svd_v = torch.svd_lowrank(adj, q=self.svd_q)
        u_mul_s = svd_u @ (torch.diag(s))
        #         v_mul_s = svd_v @ (torch.diag(s))
        recon_adj = u_mul_s @ svd_v.T
        del svd_u, s, svd_v, u_mul_s
        #         print('SVD done.')

        # 将密集矩阵转换为稀疏矩阵
        sparse_adj = recon_adj.to_sparse()
        sparse_adj = sparse_adj.to(self.device)
        # 从稀疏矩阵中提取行和列索引，这些索引代表边的连接
        row = sparse_adj.indices()[0]
        col = sparse_adj.indices()[1]
        # 创建 PyG 的 edge_index 格式
        edge_index2 = torch.stack([row, col], dim=0)
        # 从稀疏矩阵中提取边的属性，即连接的权重
        edge_weight2 = sparse_adj.values().unsqueeze(1)
        x2, y2 = data.x, data.y
        del sparse_adj

        ### original graph
        for idx, layer in enumerate(self.layers):
            x1, _ = layer(x1, edge_index1,
                          edge_attr=edge_weight1 if self.used_edge_weight else None,
                          return_attention_weights=True)  # attn1
            if self.used_DSBN:
                if self.norm:
                    if len(x1) == 1:
                        pass
                    elif self.norm[0].__class__.__name__ == 'DSBatchNorm':
                        x1 = self.norm[idx](x1, y1)
                    else:
                        x1 = self.norm[idx](x1)
                    x1 = F.relu(x1)

        z_mean1, _ = self.conv_mean(x1, edge_index1,
                                    edge_attr=edge_weight1 if self.used_edge_weight else None,
                                    return_attention_weights=True)
        z_log_std1, _ = self.conv_log_std(x1, edge_index1,
                                          edge_attr=edge_weight1 if self.used_edge_weight else None,
                                          return_attention_weights=True)

        ### SVD_adj graph
        for idx, layer in enumerate(self.layers):
            x2, _ = layer(x2, edge_index2,
                          edge_attr=edge_weight2 if self.used_edge_weight else None,
                          return_attention_weights=True)

            if self.used_DSBN:
                if self.norm:
                    if len(x2) == 1:
                        pass
                    elif self.norm[0].__class__.__name__ == 'DSBatchNorm':
                        x2 = self.norm[idx](x2, y2)
                    else:
                        x2 = self.norm[idx](x2)
                    x2 = F.relu(x2)

        z_mean2, _ = self.conv_mean(x2, edge_index2,
                                    edge_attr=edge_weight2 if self.used_edge_weight else None,
                                    return_attention_weights=True)
        z_log_std2, _ = self.conv_log_std(x2, edge_index2,
                                          edge_attr=edge_weight2 if self.used_edge_weight else None,
                                          return_attention_weights=True)

        return z_mean1, z_log_std1, z_mean2, z_log_std2


class GATDecoder(nn.Module):
    def __init__(self, in_channels, hidden_dims, out_channels, conv_type, num_heads, dropout, concat,
                 num_domains='', used_edge_weight=False, used_DSBN=False):
        """
        Initializes the GATDecoder

        Parameters
        ----------
        in_channels : int
            Number of input feature dimensions.
        hidden_dims : list[int]
            List of output dimensions for each hidden layer in the GAT, in ascending order.
        out_channels : int
            Number of output feature dimensions.
        num_heads : int
            Number of attention heads for each GAT layer.
        dropout : float
            Dropout rate for GAT layers.
        concat : bool
            Whether to concatenate outputs of all attention heads or not.
        num_domains : int or str
            Number of domains for domain-specific batch normalization (DSBN). If `1`, regular batch normalization is used.
        used_edge_weight : bool, optional
            Whether to use edge weights in the GAT layers (default is False).
        used_DSBN : bool, optional
            Whether to use domain-specific batch normalization (DSBN) (default is False).
        """
        super(GATDecoder, self).__init__()

        # self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        self.used_DSBN = used_DSBN
        self.used_edge_weight = used_edge_weight
        self.layers = nn.ModuleList()
        self.norm = nn.ModuleList()
        self.dropout = dropout
        if conv_type == 'GAT':
            GATLayer = GATConv
        elif conv_type == 'GATv2Conv':
            GATLayer = GATv2Conv

        num_hidden_layers = len(hidden_dims)
        num_heads_list = [num_heads] * num_hidden_layers
        concat_list = [concat] * num_hidden_layers
        for i in range(num_hidden_layers):
            if concat_list[i]:
                current_dim = hidden_dims[::-1][i] * num_heads_list[i]
            else:
                current_dim = hidden_dims[::-1][i]  # [::-1] 代表反转

            if type(num_domains) == int:
                if num_domains == 1:  # TO DO
                    norm = nn.BatchNorm1d(current_dim)
                else:
                    # num_domains >1 represent domain-specific batch normalization of n domain
                    norm = DSBatchNorm(current_dim, num_domains)
            else:
                norm = None
            self.norm.append(norm)

        current_dim = in_channels  # in_channels
        dropout_list = [dropout] * num_hidden_layers
        for i in range(num_hidden_layers):
            layer = GATLayer(
                in_channels=current_dim,
                out_channels=hidden_dims[::-1][i],
                heads=num_heads_list[i],
                dropout=dropout_list[i],
                concat=concat_list[i],
                edge_dim=1 if self.used_edge_weight else None
            )
            self.layers.append(layer)
            if concat_list[i]:
                current_dim = hidden_dims[::-1][i] * num_heads_list[i]
            else:
                current_dim = hidden_dims[::-1][i]

        ### 数据集总重构的layers
        self.conv_recon = GATLayer(in_channels=current_dim,
                                   out_channels=out_channels,
                                   heads=num_heads,
                                   concat=False,
                                   edge_dim=1 if self.used_edge_weight else None,
                                   dropout=dropout)

    def forward(self, x, data):
        """
        Basic block consist of:
            fc -> bn -> act -> dropout
        """
        edge_index, y, edge_weight = data.edge_index, data.y, data.edge_attr

        for idx, layer in enumerate(self.layers):
            x, _ = layer(x, edge_index, edge_attr=edge_weight if self.used_edge_weight else None,
                         return_attention_weights=True)
            if self.used_DSBN:
                if self.norm:
                    if len(x) == 1:
                        pass
                    elif self.norm[0].__class__.__name__ == 'DSBatchNorm':
                        x = self.norm[idx](x, y)
                    else:
                        x = self.norm[idx](x)
                    x = F.relu(x)

        recon_x, _ = self.conv_recon(x, edge_index,
                                     edge_attr=edge_weight if self.used_edge_weight else None,
                                     return_attention_weights=True)

        return recon_x

### GCN encoder
class GCNEncoder(nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels, svd_q, dropout=0.2,
                 num_domains='', used_edge_weight=False, used_DSBN=False):
        """
        Initializes the GCNEncoder
        Parameters
        ----------
        in_channels : int
            Number of input feature dimensions (length of each node's feature vector).
        hidden_channels : list[int]
            List of output dimensions for each hidden layer in the GCN.
        out_channels : int
            Dimension of the output feature representation produced by the encoder.
        svd_q : int
            Rank for the low-rank SVD approximation.
        dropout : float
            Dropout rate for GCN layers.
        num_domains : int or str
            Number of domains for domain-specific batch normalization (DSBN). If `1`, regular batch normalization is used.
        used_edge_weight : bool, optional
            Whether to use edge weights in the GAT layers (default is False).
        used_DSBN : bool, optional
            Whether to use domain-specific batch normalization (DSBN) (default is False).
        """
        super(GCNEncoder, self).__init__()
        # 如果 hidden_channels 是单一数字，将其转换成单元素列表
        if isinstance(hidden_channels, int):
            hidden_channels = [hidden_channels]

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.svd_q = svd_q
        self.used_DSBN = used_DSBN
        self.used_edge_weight = used_edge_weight
        self.norm = nn.ModuleList()

        # 创建一个包含所有GCN层的列表
        gcn_layers = []
        total_layers = [in_channels] + hidden_channels
        for i in range(len(total_layers) - 1):
            gcn_layers.append(GCNConv(total_layers[i], total_layers[i + 1], dropout=dropout))

        # 使用 nn.ModuleList 以确保所有层都被正确注册
        self.gcn_layers = nn.ModuleList(gcn_layers)

        # 输出层
        self.gcn_mu = GCNConv(hidden_channels[-1], out_channels, dropout=dropout)
        self.gcn_logvar = GCNConv(hidden_channels[-1], out_channels, dropout=dropout)

        # self.norm 层
        num_hidden_layers = len(hidden_channels)
        for i in range(num_hidden_layers):
            current_dim = hidden_channels[i]

            if type(num_domains) == int:
                if num_domains == 1:  # TO DO
                    norm = nn.BatchNorm1d(current_dim)
                else:
                    norm = DSBatchNorm(current_dim,
                                       num_domains)  # num_domains >1 represent domain-specific batch normalization of n domain
            else:
                norm = None
            self.norm.append(norm)

    def forward(self, data):
        x1, edge_index1, y1, edge_weight1 = data.x, data.edge_index, data.y, data.edge_attr

        # 首先，将edge_index和edge_attr转换为稀疏张量
        num_nodes = data.num_nodes  # 假设数据中节点的总数
        # 创建 SciPy 的 COO 格式稀疏矩阵
        value = torch.squeeze(data.edge_attr).cpu().numpy()
        row = data.edge_index[0].cpu().numpy()
        col = data.edge_index[1].cpu().numpy()
        sparse_adj = sp.coo_matrix((value, (row, col)),
                                   shape=(num_nodes, num_nodes))
        # perform svd reconstruction
        adj = scipy_sparse_mat_to_torch_sparse_tensor(sparse_adj).coalesce()
        svd_u, s, svd_v = torch.svd_lowrank(adj, q=self.svd_q)
        u_mul_s = svd_u @ (torch.diag(s))
        #         v_mul_s = svd_v @ (torch.diag(s))
        recon_adj = u_mul_s @ svd_v.T
        del svd_u, s, svd_v, u_mul_s

        # 将密集矩阵转换为稀疏矩阵
        sparse_adj = recon_adj.to_sparse()
        sparse_adj = sparse_adj.to(self.device)
        # 从稀疏矩阵中提取行和列索引，这些索引代表边的连接
        row = sparse_adj.indices()[0]
        col = sparse_adj.indices()[1]
        # 创建 PyG 的 edge_index 格式
        edge_index2 = torch.stack([row, col], dim=0)
        # 从稀疏矩阵中提取边的属性，即连接的权重
        edge_weight2 = sparse_adj.values().unsqueeze(1)
        x2, y2 = data.x, data.y
        del sparse_adj

        ### original graph
        for idx, layer in enumerate(self.gcn_layers):
            x1 = layer(x1, edge_index1,
                       edge_weight=edge_weight1 if self.used_edge_weight else None)

            if self.used_DSBN:
                if self.norm:
                    if len(x1) == 1:
                        pass
                    elif self.norm[0].__class__.__name__ == 'DSBatchNorm':
                        x1 = self.norm[idx](x1, y1)
                    else:
                        x1 = self.norm[idx](x1)
                    x1 = F.relu(x1)

        z_mean1 = self.gcn_mu(x1, edge_index1,
                              edge_weight=edge_weight1 if self.used_edge_weight else None)
        z_log_std1 = self.gcn_logvar(x1, edge_index1,
                                     edge_weight=edge_weight1 if self.used_edge_weight else None)

        ### SVD_adj graph
        for idx, layer in enumerate(self.gcn_layers):
            x2 = layer(x2, edge_index2,
                       edge_weight=edge_weight2 if self.used_edge_weight else None)

            if self.used_DSBN:
                if self.norm:
                    if len(x2) == 1:
                        pass
                    elif self.norm[0].__class__.__name__ == 'DSBatchNorm':
                        x2 = self.norm[idx](x2, y2)
                    else:
                        x2 = self.norm[idx](x2)
                    x2 = F.relu(x2)

        z_mean2 = self.gcn_mu(x2, edge_index2,
                              edge_weight=edge_weight2 if self.used_edge_weight else None)
        z_log_std2 = self.gcn_logvar(x2, edge_index2,
                                     edge_weight=edge_weight2 if self.used_edge_weight else None)

        return z_mean1, z_log_std1, z_mean2, z_log_std2


### GCN decoder
class GCNDecoder(nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels, dropout=0.2,
                 num_domains='', used_edge_weight=False, used_DSBN=False):
        """
        Initializes the GCNDecoder

        Parameters
        ----------
        in_channels : int
            Number of input feature dimensions.
        hidden_channels : list[int]
            List of output dimensions for each hidden layer in the GCN, in ascending order.
        out_channels : int
            Number of output feature dimensions.
        dropout : float
            Dropout rate for GAT layers.
        num_domains : int or str
            Number of domains for domain-specific batch normalization (DSBN). If `1`, regular batch normalization is used.
        used_edge_weight : bool, optional
            Whether to use edge weights in the GAT layers (default is False).
        used_DSBN : bool, optional
            Whether to use domain-specific batch normalization (DSBN) (default is False).
        """
        super(GCNDecoder, self).__init__()
        # 如果 hidden_channels 是单一数字，将其转换成单元素列表
        if isinstance(hidden_channels, int):
            hidden_channels = [hidden_channels]

        self.used_DSBN = used_DSBN
        self.used_edge_weight = used_edge_weight
        self.layers = nn.ModuleList()
        self.norm = nn.ModuleList()

        # 创建一个包含所有GCN层的列表
        gcn_layers = []
        hidden_channels = hidden_channels[::-1]  # 反转
        total_layers = [in_channels] + hidden_channels
        for i in range(len(total_layers) - 1):
            gcn_layers.append(GCNConv(total_layers[i], total_layers[i + 1], dropout=dropout))

        # 使用 nn.ModuleList 以确保所有层都被正确注册
        self.gcn_layers = nn.ModuleList(gcn_layers)

        # 输出层
        self.gcn_recon = GCNConv(hidden_channels[-1], out_channels, dropout=dropout)

        # self.norm 层
        num_hidden_layers = len(hidden_channels)
        for i in range(num_hidden_layers):
            current_dim = hidden_channels[i]

            if type(num_domains) == int:
                if num_domains == 1:  # TO DO
                    norm = nn.BatchNorm1d(current_dim)
                else:
                    norm = DSBatchNorm(current_dim,
                                       num_domains)  # num_domains >1 represent domain-specific batch normalization of n domain
            else:
                norm = None
            self.norm.append(norm)

    def forward(self, x, data):
        edge_index, y, edge_weight = data.edge_index, data.y, data.edge_attr

        ### latent
        for idx, layer in enumerate(self.gcn_layers):
            x = layer(x, edge_index,
                      edge_weight=edge_weight if self.used_edge_weight else None)

            if self.used_DSBN:
                if self.norm:
                    if len(x) == 1:
                        pass
                    elif self.norm[0].__class__.__name__ == 'DSBatchNorm':
                        x = self.norm[idx](x, y)
                    else:
                        x = self.norm[idx](x)
                    x = F.relu(x)

        x_recon = self.gcn_recon(x, edge_index,
                                 edge_weight=edge_weight if self.used_edge_weight else None)

        return x_recon


# Based on VGAE class in PyTorch Geometric
EPS = 1e-15
MAX_LOGSTD = 10
class GNNModelVAE(GAE):
    def __init__(self, encoder, bottle_neck_neurons, hidden_dims, feature_dim,
                 num_heads, dropout, concat, n_domain, used_edge_weight, used_DSBN,
                 used_recon_exp, conv_type, cluster_num, gnn_layer=2, decoder=None):
        """
        Initializes the GCNModelVAE
        Parameters
        ----------
        encoder : nn.Module
            The encoder module used in the variational graph autoencoder. 'GAT' or 'GCN'
        enc_in_channels : int
            Number of input feature dimensions for the encoder.
        n_domain : int
            Number of domains for domain-specific batch normalization (DSBN).
        args : Namespace
            Arguments that include hyperparameters and configurations such as dropout rate, latent dimensions, etc.
        decoder : nn.Module, optional
            The decoder module used in the variational autoencoder. Defaults to `InnerProductDecoder` if not specified.
        """
        super(GNNModelVAE, self).__init__(encoder, decoder)
        self.decoder = InnerProductDecoder() if decoder is None else decoder

        self.latent = bottle_neck_neurons
        self.hidden_dims = hidden_dims
        self.feature_dim = feature_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.concat = concat
        self.n_domain = n_domain
        self.used_edge_weight = used_edge_weight
        self.used_DSBN = used_DSBN

        # model configurations
        self.used_recon_exp = used_recon_exp
        self.conv_type = conv_type
        assert self.conv_type in ['GAT', 'GATv2Conv', 'GCN'], 'Convolution must be "GCN", "GAT" or "GATv2Conv".'
        self.cluster_num = cluster_num
        self.gnn_layer = gnn_layer
        # 使用 Xavier 初始化权重
        self.eps_weight = nn.Parameter(nn.init.xavier_uniform_(torch.empty(self.latent, self.latent)))
        self.eps_bias = nn.Parameter(torch.zeros(self.latent))
        self.instance_projector = nn.Sequential(
            nn.Linear(self.latent, self.latent),
            nn.LayerNorm(self.latent, elementwise_affine=False),
            nn.ReLU(),
            nn.Linear(self.latent, self.feature_dim),
        )
        self.cluster_projector = nn.Sequential(
            nn.Linear(self.latent, self.latent),
            nn.LayerNorm(self.latent, elementwise_affine=False),
            nn.ReLU(),
            nn.Linear(self.latent, self.cluster_num),
            nn.Softmax(dim=1)
        )

        ## 重构表达谱
        if self.used_recon_exp:
            if self.conv_type in ['GAT', 'GATv2Conv']:
                self.GAT_decoder = GATDecoder(
                    in_channels=self.latent,
                    hidden_dims=self.hidden_dims,
                    out_channels=self.feature_dim,
                    conv_type=self.conv_type,
                    num_heads=self.num_heads,
                    dropout=self.dropout,
                    concat=self.concat,
                    num_domains=self.n_domain,  # DSBN
                    used_edge_weight=self.used_edge_weight,
                    used_DSBN=self.used_DSBN
                )
            elif self.conv_type == 'GCN':
                self.GCN_decoder = GCNDecoder(
                    in_channels=self.latent,
                    hidden_channels=self.hidden_dims,
                    out_channels=self.feature_dim,
                    dropout=self.dropout,
                    num_domains=self.n_domain,  # DSBN
                    used_edge_weight=self.used_edge_weight,
                    used_DSBN=self.used_DSBN
                )
            else:
                raise NotImplementedError("Unknown GNN-Operator.")

    def reparametrize(self, mu: Tensor, logstd: Tensor, eps=None) -> Tensor:
        if self.training:
            if eps is not None:
                return mu + eps * torch.randn_like(logstd) * torch.exp(logstd)
            else:
                return mu + torch.randn_like(logstd) * torch.exp(logstd)
        else:
            return mu

    def encode(self, *args, **kwargs) -> Tensor:
        """"""  # noqa: D419
        if self.conv_type in ['GAT', 'GATv2Conv']:
            self.__mu1__, self.__logstd1__, self.__mu2__, self.__logstd2__ = self.encoder(*args, **kwargs)
            return self.__mu1__, self.__logstd1__, self.__mu2__, self.__logstd2__
        else:
            self.__mu1__, self.__logstd1__, self.__mu2__, self.__logstd2__ = self.encoder(*args, **kwargs)
            return self.__mu1__, self.__logstd1__, self.__mu2__, self.__logstd2__

    def recon_loss(self, z: Tensor, pos_edge_index: Tensor,
                   neg_edge_index: Optional[Tensor] = None) -> Tensor:
        r"""Given latent variables :obj:`z`, computes the binary cross
        entropy loss for positive edges :obj:`pos_edge_index` and negative
        sampled edges.

        Args:
            z (torch.Tensor): The latent space :math:`\mathbf{Z}`.
            pos_edge_index (torch.Tensor): The positive edges to train against.
            neg_edge_index (torch.Tensor, optional): The negative edges to
                train against. If not given, uses negative sampling to
                calculate negative edges. (default: :obj:`None`)
        """
        pos_loss = -torch.log(
            self.decoder(z, pos_edge_index, sigmoid=True) + EPS).mean()

        # Do not include self-loops in negative samples
        pos_edge_index, _ = remove_self_loops(pos_edge_index)
        pos_edge_index, _ = add_self_loops(pos_edge_index)

        if neg_edge_index is None:
            neg_edge_index = negative_sampling(pos_edge_index, z.size(0))
        neg_loss = -torch.log(1 -
                              self.decoder(z, neg_edge_index.long(), sigmoid=True) +
                              EPS).mean()

        return pos_loss + neg_loss

    def kl_loss(self, mu: Optional[Tensor] = None,
                logstd: Optional[Tensor] = None) -> Tensor:
        r"""Computes the KL loss, either for the passed arguments :obj:`mu`
        and :obj:`logstd`, or based on latent variables from last encoding.

        Args:
            mu (torch.Tensor, optional): The latent space for :math:`\mu`. If
                set to :obj:`None`, uses the last computation of :math:`\mu`.
                (default: :obj:`None`)
            logstd (torch.Tensor, optional): The latent space for
                :math:`\log\sigma`.  If set to :obj:`None`, uses the last
                computation of :math:`\log\sigma^2`. (default: :obj:`None`)
        """
        mu = self.__mu1__ if mu is None else mu
        logstd = self.__logstd1__ if logstd is None else logstd.clamp(
            max=MAX_LOGSTD)
        return -0.5 * torch.mean(
            torch.sum(1 + 2 * logstd - mu ** 2 - logstd.exp() ** 2, dim=1))

    def calc_individual_metrics(self, z: Tensor, pos_edge_index: Tensor,
                                neg_edge_index: Tensor) -> Tuple[Tensor, Tensor]:
        r"""Given latent variables :obj:`z`, positive edges
        :obj:`pos_edge_index` and negative edges :obj:`neg_edge_index`,
        computes area under the ROC curve (AUC) and average precision (AP)
        scores.

        Args:
            z (torch.Tensor): The latent space :math:`\mathbf{Z}`.
            pos_edge_index (torch.Tensor): The positive edges to evaluate
                against.
            neg_edge_index (torch.Tensor): The negative edges to evaluate
                against.
        """
        from sklearn.metrics import roc_auc_score, average_precision_score, accuracy_score, f1_score, roc_curve, \
            precision_recall_curve, silhouette_score, calinski_harabasz_score, davies_bouldin_score

        pos_y = z.new_ones(pos_edge_index.size(1))
        neg_y = z.new_zeros(neg_edge_index.size(1))
        y = torch.cat([pos_y, neg_y], dim=0)

        pos_pred = self.decoder(z, pos_edge_index, sigmoid=True)
        neg_pred = self.decoder(z, neg_edge_index, sigmoid=True)
        pred = torch.cat([pos_pred, neg_pred], dim=0)

        y, pred = y.detach().cpu().numpy(), pred.detach().cpu().numpy()

        try:
            roc_score = roc_auc_score(y, pred)
        except ValueError:
            roc_score = 0.5
        ap_score = average_precision_score(y, pred)
        acc = accuracy_score(y, pred > 0.5)
        f1 = f1_score(y, pred > 0.5, average='micro')

        return roc_score, ap_score, acc, f1

    def encodeBatch(self, data):
        all_mu1 = []
        for _ in range(self.gnn_layer):
            if self.conv_type in ['GAT', 'GATv2Conv']:
                mu1, _, _, _ = self.encode(data)
                all_mu1.append(mu1)
            else:
                mu1, _, _, _ = self.encode(data)
                all_mu1.append(mu1)

        mean1 = torch.stack(all_mu1).mean(dim=0)  # sum
        logstd1 = torch.matmul(mean1, self.eps_weight) + self.eps_bias
        z1 = self.reparametrize(mean1, logstd1)

        ## 重构表达矩阵
        if self.used_recon_exp:
            if self.conv_type in ['GAT', 'GATv2Conv']:
                recon_features = self.GAT_decoder(z1, data)
            else:
                recon_features = self.GCN_decoder(z1, data)
            return mean1, recon_features
        else:
            return mean1

    def forward(self, data):
        all_mu1 = []
        all_mu2 = []
        for _ in range(self.gnn_layer):
            if self.conv_type in ['GAT', 'GATv2Conv']:
                mu1, _, mu2, _ = self.encode(data)
                all_mu1.append(mu1)
                all_mu2.append(mu2)
            else:
                mu1, _, mu2, _ = self.encode(data)
                all_mu1.append(mu1)
                all_mu2.append(mu2)

        mean1 = torch.stack(all_mu1).mean(dim=0)  # sum
        logstd1 = torch.matmul(mean1, self.eps_weight) + self.eps_bias
        z1 = self.reparametrize(mean1, logstd1)

        mean2 = torch.stack(all_mu2).mean(dim=0)
        logstd2 = torch.matmul(mean2, self.eps_weight) + self.eps_bias
        z2 = self.reparametrize(mean2, logstd2)

        z_1 = normalize(self.instance_projector(z1), dim=1)
        z_2 = normalize(self.instance_projector(z2), dim=1)

        c_1 = self.cluster_projector(z1)
        c_2 = self.cluster_projector(z2)

        # adj_pred = self.decoder.forward_all(z, sigmoid=False)

        ## 重构表达矩阵
        if self.used_recon_exp:
            if self.conv_type in ['GAT', 'GATv2Conv']:
                recon_features = self.GAT_decoder(z1, data)
            else:
                recon_features = self.GCN_decoder(z1, data)
            return z1, z_1, z_2, c_1, c_2, mean1, logstd1, recon_features
        else:
            return z1, z_1, z_2, c_1, c_2, mean1, logstd1

    def single_test(self, data):
        # Ensure test_data contains neg_edge_label_index
        if not hasattr(data, 'neg_edge_label_index'):
            data.neg_edge_label_index = negative_sampling(
                edge_index=data.pos_edge_label_index,
                num_nodes=data.num_nodes
            )

        pos_edge_label_index, neg_edge_label_index = data.pos_edge_label_index.long(), data.neg_edge_label_index.long()
        with torch.no_grad():
            all_mu = []
            for _ in range(self.gnn_layer):
                if self.conv_type in ['GAT', 'GATv2Conv']:
                    mu1, _, _, _ = self.encode(data)
                    all_mu.append(mu1)
                else:
                    mu1, _, _, _ = self.encode(data)
                    all_mu.append(mu1)

            mean = torch.stack(all_mu).mean(dim=0)
            logstd = torch.matmul(mean, self.eps_weight) + self.eps_bias
            z = self.reparametrize(mean, logstd)

        roc_score, ap_score, acc, f1 = self.calc_individual_metrics(z, pos_edge_label_index, neg_edge_label_index)
        return roc_score, ap_score, acc, f1


"""Garfield model"""
from .prepare_Data import (
    UserDataset
)
from .GarfieldTrainer import (
    GarfieldTrainer
)
from .Garfield_net import (
    Garfield
)
from ._layers import (
    DSBatchNorm,
    GATEncoder,
    GATDecoder,
    GCNEncoder,
    GCNDecoder,
    GNNModelVAE
)
from ._loss import (
    VGAE_loss,
    InstanceLoss,
    ClusterLoss,
    mmd_loss_calc
)
from .metrics import (
    batch_entropy_mixing_score,
    silhouette
)
from ._tools import (
    EarlyStopping,
    print_progress
)
from ._utils import (
    Transfer_scData,
    scipy_sparse_mat_to_torch_sparse_tensor
)



import scanpy as sc
import pandas as pd
import os
import numpy as np
import anndata
from sklearn.neighbors import KNeighborsTransformer

'''
Author: Lisa Sikemma
Modified from scArches
'''

def weighted_knn_trainer(train_adata, train_adata_emb, n_neighbors=50):
    """Trains a weighted KNN classifier on ``train_adata``.
    Parameters
    ----------
    train_adata: :class:`~anndata.AnnData`
        Annotated dataset to be used to train KNN classifier with ``label_key`` as the target variable.
    train_adata_emb: str
        Name of the obsm layer to be used for calculation of neighbors. If set to "X", anndata.X will be
        used
    n_neighbors: int
        Number of nearest neighbors in KNN classifier.
    """
    print(
        f"Weighted KNN with n_neighbors = {n_neighbors} ... "
    )
    k_neighbors_transformer = KNeighborsTransformer(
        n_neighbors=n_neighbors,
        mode="distance",
        algorithm="brute",
        metric="euclidean",
        n_jobs=-1,
    )
    if train_adata_emb == "X":
        train_emb = train_adata.X
    elif train_adata_emb in train_adata.obsm.keys():
        train_emb = train_adata.obsm[train_adata_emb]
    else:
        raise ValueError(
            "train_adata_emb should be set to either 'X' or the name of the obsm layer to be used!"
        )
    k_neighbors_transformer.fit(train_emb)
    return k_neighbors_transformer

def weighted_knn_transfer(
    query_adata,
    query_adata_emb,
    ref_adata_obs,
    label_keys,
    knn_model,
    threshold=1,
    pred_unknown=False,
    mode="package",
):
    """Annotates ``query_adata`` cells with an input trained weighted KNN classifier.
    Parameters
    ----------
    query_adata: :class:`~anndata.AnnData`
        Annotated dataset to be used to queryate KNN classifier. Embedding to be used
    query_adata_emb: str
        Name of the obsm layer to be used for label transfer. If set to "X",
        query_adata.X will be used
    ref_adata_obs: :class:`pd.DataFrame`
        obs of ref Anndata
    label_keys: str
        Names of the columns to be used as target variables (e.g. cell_type) in ``query_adata``.
    knn_model: :class:`~sklearn.neighbors._graph.KNeighborsTransformer`
        knn model trained on reference adata with weighted_knn_trainer function
    threshold: float
        Threshold of uncertainty used to annotating cells as "Unknown". cells with
        uncertainties higher than this value will be annotated as "Unknown".
        Set to 1 to keep all predictions. This enables one to later on play
        with thresholds.
    pred_unknown: bool
        ``False`` by default. Whether to annotate any cell as "unknown" or not.
        If `False`, ``threshold`` will not be used and each cell will be annotated
        with the label which is the most common in its ``n_neighbors`` nearest cells.
    mode: str
        Has to be one of "paper" or "package". If mode is set to "package",
        uncertainties will be 1 - P(pred_label), otherwise it will be 1 - P(true_label).
    """
    if not type(knn_model) == KNeighborsTransformer:
        raise ValueError(
            "knn_model should be of type sklearn.neighbors._graph.KNeighborsTransformer!"
        )

    if query_adata_emb == "X":
        query_emb = query_adata.X
    elif query_adata_emb in query_adata.obsm.keys():
        query_emb = query_adata.obsm[query_adata_emb]
    else:
        raise ValueError(
            "query_adata_emb should be set to either 'X' or the name of the obsm layer to be used!"
        )
    top_k_distances, top_k_indices = knn_model.kneighbors(X=query_emb)

    stds = np.std(top_k_distances, axis=1)
    stds = (2.0 / stds) ** 2
    stds = stds.reshape(-1, 1)

    top_k_distances_tilda = np.exp(-np.true_divide(top_k_distances, stds))

    weights = top_k_distances_tilda / np.sum(
        top_k_distances_tilda, axis=1, keepdims=True
    )
    cols = ref_adata_obs.columns[ref_adata_obs.columns.str.startswith(label_keys)]
    uncertainties = pd.DataFrame(columns=cols, index=query_adata.obs_names)
    pred_labels = pd.DataFrame(columns=cols, index=query_adata.obs_names)
    for i in range(len(weights)):
        for j in cols:
            y_train_labels = ref_adata_obs[j].values
            unique_labels = np.unique(y_train_labels[top_k_indices[i]])
            best_label, best_prob = None, 0.0
            for candidate_label in unique_labels:
                candidate_prob = weights[
                    i, y_train_labels[top_k_indices[i]] == candidate_label
                ].sum()
                if best_prob < candidate_prob:
                    best_prob = candidate_prob
                    best_label = candidate_label

            if pred_unknown:
                if best_prob >= threshold:
                    pred_label = best_label
                else:
                    pred_label = "Unknown"
            else:
                pred_label = best_label

            if mode == "package":
                uncertainties.iloc[i][j] = (max(1 - best_prob, 0))

            else:
                raise Exception("Inquery Mode!")

            pred_labels.iloc[i][j] = (pred_label)

    print("Label transfer finished!", flush=True)

    return pred_labels, uncertainties


import torch
import torch.nn as nn
import math
from torch.autograd import Variable

### VGAE loss
def VGAE_loss(VGAE_model, z, mu, logstd, pos_edge_index, neg_edge_index=None):
    """
    Computes the variational graph autoencoder (VGAE) loss.

    Parameters
    ----------
    VGAE_model : torch.nn.Module
        The VGAE model instance used to compute the reconstruction and KL divergence losses.
    z : torch.Tensor
        The latent space representation :math:`\mathbf{Z}`.
    mu : torch.Tensor
        The mean of the latent space distribution.
    logstd : torch.Tensor
        The logarithm of the standard deviation of the latent space distribution.
    pos_edge_index : torch.Tensor
        Indices of positive edges to train against, in COO format.
    neg_edge_index : torch.Tensor, optional
        Indices of negative edges to train against, in COO format. If not provided, negative sampling is used.
        Defaults to `None`.

    Returns
    -------
    torch.Tensor
        The sum of reconstruction loss and KL divergence loss.
    """
    rec_loss = VGAE_model.recon_loss(z, pos_edge_index, neg_edge_index)
    kl_loss = 1 / z.size(0) * VGAE_model.kl_loss(mu, logstd)

    return rec_loss + kl_loss

## contrastive loss
class InstanceLoss(nn.Module):
    def __init__(self, temperature, device):
        super(InstanceLoss, self).__init__()
        self.temperature = temperature
        self.criterion = nn.CrossEntropyLoss(reduction="sum")

    def mask_correlated_samples(self, batch_size):
        N = 2 * batch_size
        mask = torch.ones((N, N))
        mask = mask.fill_diagonal_(0)
        for i in range(batch_size):
            mask[i, batch_size + i] = 0
            mask[batch_size + i, i] = 0
        mask = mask.bool()
        return mask

    def forward(self, z_i, z_j):
        batch_size = z_i.size(0)
        mask = self.mask_correlated_samples(batch_size)
        N = 2 * batch_size
        z = torch.cat((z_i, z_j), dim=0)

        sim = torch.matmul(z, z.T) / self.temperature
        sim_i_j = torch.diag(sim, batch_size)
        sim_j_i = torch.diag(sim, -batch_size)

        positive_samples = torch.cat((sim_i_j, sim_j_i), dim=0).reshape(N, 1)
        negative_samples = sim[mask].reshape(N, -1)

        labels = torch.zeros(N).to(positive_samples.device).long()
        logits = torch.cat((positive_samples, negative_samples), dim=1)
        loss = self.criterion(logits, labels)
        loss /= N

        return loss

class ClusterLoss(nn.Module):
    def __init__(self, class_num, temperature, device):
        super(ClusterLoss, self).__init__()
        self.class_num = class_num
        self.temperature = temperature

        self.mask = self.mask_correlated_clusters(class_num)
        self.criterion = nn.CrossEntropyLoss(reduction="sum")
        self.similarity_f = nn.CosineSimilarity(dim=2)

    def mask_correlated_clusters(self, class_num):
        N = 2 * class_num
        mask = torch.ones((N, N))
        mask = mask.fill_diagonal_(0)
        for i in range(class_num):
            mask[i, class_num + i] = 0
            mask[class_num + i, i] = 0
        mask = mask.bool()
        return mask

    def forward(self, c_i, c_j):
        p_i = c_i.sum(0).view(-1)
        p_i /= p_i.sum()
        ne_i = math.log(p_i.size(0)) + (p_i * torch.log(p_i)).sum()
        p_j = c_j.sum(0).view(-1)
        p_j /= p_j.sum()
        ne_j = math.log(p_j.size(0)) + (p_j * torch.log(p_j)).sum()
        ne_loss = ne_i + ne_j

        c_i = c_i.t()
        c_j = c_j.t()
        N = 2 * self.class_num
        c = torch.cat((c_i, c_j), dim=0)

        sim = self.similarity_f(c.unsqueeze(1), c.unsqueeze(0)) / self.temperature
        sim_i_j = torch.diag(sim, self.class_num)
        sim_j_i = torch.diag(sim, -self.class_num)

        positive_clusters = torch.cat((sim_i_j, sim_j_i), dim=0).reshape(N, 1)
        negative_clusters = sim[self.mask].reshape(N, -1)

        labels = torch.zeros(N).to(positive_clusters.device).long()
        logits = torch.cat((positive_clusters, negative_clusters), dim=1)
        loss = self.criterion(logits, labels)
        loss /= N

        return loss + ne_loss

### mmd function
def pairwise_distance(x, y):
    x = x.view(x.shape[0], x.shape[1], 1)
    y = torch.transpose(y, 0, 1)
    output = torch.sum((x - y) ** 2, 1)
    output = torch.transpose(output, 0, 1)

    return output

def gaussian_kernel_matrix(x, y, alphas):
    """Computes multiscale-RBF kernel between x and y.

       Parameters
       ----------
       x: torch.Tensor
            Tensor with shape [batch_size, z_dim].
       y: torch.Tensor
            Tensor with shape [batch_size, z_dim].
       alphas: Tensor

       Returns
       -------
       Returns the computed multiscale-RBF kernel between x and y.
    """

    dist = pairwise_distance(x, y).contiguous()
    dist_ = dist.view(1, -1)

    alphas = alphas.view(alphas.shape[0], 1)
    beta = 1. / (2. * alphas)

    s = torch.matmul(beta, dist_)

    return torch.sum(torch.exp(-s), 0).view_as(dist)


def mmd_loss_calc(source_features, target_features):
    """Initializes Maximum Mean Discrepancy(MMD) between source_features and target_features.

       - Gretton, Arthur, et al. "A Kernel Two-Sample Test". 2012.

       Parameters
       ----------
       source_features: torch.Tensor
            Tensor with shape [batch_size, z_dim]
       target_features: torch.Tensor
            Tensor with shape [batch_size, z_dim]

       Returns
       -------
       Returns the computed MMD between x and y.
    """
    alphas = [
        1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1, 5, 10, 15, 20, 25, 30, 35, 100,
        1e3, 1e4, 1e5, 1e6
    ]
    alphas = Variable(torch.FloatTensor(alphas)).to(device=source_features.device)

    cost = torch.mean(gaussian_kernel_matrix(source_features, source_features, alphas))
    cost += torch.mean(gaussian_kernel_matrix(target_features, target_features, alphas))
    cost -= 2 * torch.mean(gaussian_kernel_matrix(source_features, target_features, alphas))

    return cost





"""Quality Control"""
import numpy as np
from scipy.sparse import (
    issparse,
    csr_matrix,
)
import re

def cal_qc_rna(adata, expr_cutoff=1):
    """Calculate quality control metrics.

    Parameters
    ----------
    adata: AnnData
        Annotated data matrix.
    expr_cutoff: `float`, optional (default: 1)
        Expression cutoff.
        If greater than expr_cutoff,the feature is considered 'expressed'
    assay: `str`, optional (default: 'rna')
            Choose from {'rna','atac'},case insensitive
    Returns
    -------
    updates `adata` with the following fields.
    n_counts: `pandas.Series` (`adata.var['n_counts']`,dtype `int`)
       The number of read count each gene has.
    n_cells: `pandas.Series` (`adata.var['n_cells']`,dtype `int`)
       The number of cells in which each gene is expressed.
    pct_cells: `pandas.Series` (`adata.var['pct_cells']`,dtype `float`)
       The percentage of cells in which each gene is expressed.
    n_counts: `pandas.Series` (`adata.obs['n_counts']`,dtype `int`)
       The number of read count each cell has.
    n_genes: `pandas.Series` (`adata.obs['n_genes']`,dtype `int`)
       The number of genes expressed in each cell.
    pct_genes: `pandas.Series` (`adata.obs['pct_genes']`,dtype `float`)
       The percentage of genes expressed in each cell.
    n_peaks: `pandas.Series` (`adata.obs['n_peaks']`,dtype `int`)
       The number of peaks expressed in each cell.
    pct_peaks: `pandas.Series` (`adata.obs['pct_peaks']`,dtype `int`)
       The percentage of peaks expressed in each cell.
    pct_mt: `pandas.Series` (`adata.obs['pct_mt']`,dtype `float`)
       the percentage of counts in mitochondrial genes
    """

    if(not issparse(adata.X)):
        adata.X = csr_matrix(adata.X)

    n_counts = adata.X.sum(axis=0).A1
    adata.var['n_counts'] = n_counts
    n_cells = (adata.X >= expr_cutoff).sum(axis=0).A1
    adata.var['n_cells'] = n_cells
    adata.var['pct_cells'] = n_cells/adata.shape[0]

    n_counts = adata.X.sum(axis=1).A1
    adata.obs['n_counts'] = n_counts
    n_features = (adata.X >= expr_cutoff).sum(axis=1).A1
    adata.obs['n_genes'] = n_features
    adata.obs['pct_genes'] = n_features/adata.shape[1]
    r = re.compile("^MT-", flags=re.IGNORECASE)
    mt_genes = list(filter(r.match, adata.var_names))
    if(len(mt_genes) > 0):
        n_counts_mt = adata[:, mt_genes].X.sum(axis=1).A1
        adata.obs['pct_mt'] = n_counts_mt/n_counts
    else:
        adata.obs['pct_mt'] = 0
